"""Filter function helpers."""

from __future__ import annotations

import collections.abc

from abc import ABC, abstractmethod
from functools import wraps

from typing import Tuple
from typing import Any
from typing import TYPE_CHECKING

from liquid.exceptions import FilterArgumentError

if TYPE_CHECKING:
    from liquid import Environment


class Filter(ABC):
    """Base class for all filters.

    If with_context is True, the current render context will be passed as a
    keyword argument every time the filter is called.
    """

    __slots__ = ("env",)

    with_context: bool = False
    name: str = "filter"

    def __init__(self, env: Environment):
        self.env = env

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """Call the filter where `val` is the left hand expression of an output
        statement."""


def expect_number(name: str, val: object):
    """Return `val` as an integer or float. Raise a FilterArgumentError
    if `val` can't be cast to an integer or float."""
    if isinstance(val, (int, float)):
        num = val
    elif isinstance(val, str):
        if val.isdecimal():
            num = int(val)
        try:
            num = float(val)
        except ValueError as err:
            raise FilterArgumentError(
                f"{name}: could not cast string '{val}' to a number"
            ) from err
    else:
        raise FilterArgumentError(
            f"expected an int, float or string, found {type(val)}"
        )
    return num


def expect_integer(name: str, val: object):
    """Return `val` as an integer. Raise a FilterArgumentError if `val` can't
    be cast to an integer."""
    if isinstance(val, int):
        num = val
    elif isinstance(val, str):
        if val.isdecimal():
            num = int(val)
        else:
            raise FilterArgumentError(
                f"{name}: could not cast string '{val}' to an integer"
            )
    else:
        raise FilterArgumentError(f"expected an int or string, found {type(val)}")
    return num


def expect_string(name: str, val: object):
    """Raise a FilterArgumentError is val is not a string."""
    if not isinstance(val, str):
        raise FilterArgumentError(f"{name}: expected a string, found {type(val)}")


def expect_string_or_array(name: str, val: object):
    """Raise a FilterArgumentError is val is not a string or a list."""
    if not isinstance(val, (str, list)):
        raise FilterArgumentError(
            f"{name}: expected a string or array, found {type(val)}"
        )


def expect_n_args(name: str, n: int, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args does not have `n` elements, and
    kwargs is not empty."""
    length = len(args)
    if length != n:
        raise FilterArgumentError(
            f"{name} expects a one positional argument, found {length}"
        )
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )


def expect_one_arg(name: str, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args is not a single element tuple, and
    kwargs is not empty."""
    expect_n_args(name, 1, args, kwargs)


def maybe_one_arg(name: str, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args is not empty or a single element tuple,
    and kwargs is not empty."""
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )

    length = len(args)
    if length == 1:
        return True
    if length == 0:
        return False

    raise FilterArgumentError(
        f"{name}: expected zero or one positional arguments, found {length}"
    )


def one_maybe_two_args(name: str, args: ..., kwargs: ...) -> Tuple[Any, ...]:
    """Return (args[0], args[1]) or None if the args aren't there."""
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )

    length = len(args)
    if length == 2:
        return args[0], args[1]
    if length == 1:
        return args[0], None

    raise FilterArgumentError(
        f"{name}: expected one or two positional arguments, found {length}"
    )


def expect_no_args(name: str, args: ..., kwargs: ...):
    """Raise a FilterArgumentError if args and kwargs are not empty."""
    if args:
        raise FilterArgumentError(
            f"{name}: expected zero positional arguments, found {len(args)}"
        )
    if kwargs:
        raise FilterArgumentError(
            f"{name}: expected zero keyword arguments, found {len(kwargs)}"
        )


def expect_array(name: str, val: object):
    """Raise a FilterArgumentError is val is not an array/list."""
    if not isinstance(val, (list, tuple)):
        raise FilterArgumentError(f"{name}: expected an array, found {type(val)}")


def array_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_array(instance.name, val)
        return func(instance, val, *args, **kwargs)

    return wrapper


def array_or_string_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_string_or_array(instance.name, val)
        return func(instance, val, *args, **kwargs)

    return wrapper


def array_of_strings_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_array(instance.name, val)
        for item in val:
            if not isinstance(item, str):
                raise FilterArgumentError(
                    f"{instance.name}: expected an array of strings, found a {type(item)}"
                )
        return func(instance, val, *args, **kwargs)

    return wrapper


def array_of_hashable_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_array(instance.name, val)
        for item in val:
            if not isinstance(item, collections.abc.Hashable):
                raise FilterArgumentError(
                    f"{instance.name}: expected an array of hashable items, "
                    f"found a {type(item)}"
                )
        return func(instance, val, *args, **kwargs)

    return wrapper


def string_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_string(instance.name, val)
        return func(instance, val, *args, **kwargs)

    return wrapper


def number_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        num = expect_number(instance.name, val)
        return func(instance, num, *args, **kwargs)

    return wrapper


def one_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        return func(instance, val, args[0], **kwargs)

    return wrapper


def maybe_one_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        maybe_one_arg(instance.name, args, kwargs)
        return func(instance, val, *args, **kwargs)

    return wrapper


def one_maybe_two_args_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        one_maybe_two_args(instance.name, args, kwargs)
        return func(instance, val, *args, **kwargs)

    return wrapper


def one_array_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        expect_array(instance.name, args[0])
        return func(instance, val, args[0], **kwargs)

    return wrapper


def one_number_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        num = expect_number(instance.name, args[0])
        return func(instance, val, num, **kwargs)

    return wrapper


def one_string_arg_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_one_arg(instance.name, args, kwargs)
        expect_string(instance.name, args[0])
        return func(instance, val, args[0], **kwargs)

    return wrapper


def two_string_args_required(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_n_args(instance.name, 2, args, kwargs)
        expect_string(instance.name, args[0])
        expect_string(instance.name, args[1])
        return func(instance, val, args[0], args[1], **kwargs)

    return wrapper


def no_args(func):
    @wraps(func)
    def wrapper(instance, val, *args, **kwargs):
        expect_no_args(instance.name, args, kwargs)
        return func(instance, val, *args, **kwargs)

    return wrapper



import warnings

from enum import IntEnum, auto

from typing import Union
from typing import Optional
from typing import Type

from liquid.exceptions import Error
from liquid.exceptions import lookup_warning


class Mode(IntEnum):
    """Template correctness tolerance."""

    LAX = auto()
    WARN = auto()
    STRICT = auto()


def error(
    tolerance: Mode,
    exc: Union[Type[Error], Error],
    msg: Optional[str] = None,
    linenum: Optional[int] = None,
):
    """Raise an exception or warning according to the given mode."""
    if not isinstance(exc, Error):
        exc = exc(msg, linenum=linenum)
    elif not exc.linenum:
        exc.linenum = linenum

    if tolerance == Mode.STRICT:
        raise exc
    if tolerance == Mode.WARN:
        warnings.warn(str(exc), category=lookup_warning(exc.__class__))



"""Test miscellaneous filter functions."""
import datetime
import unittest
from inspect import isclass
from typing import NamedTuple, Any, List, Dict

from liquid.environment import Environment
from liquid.exceptions import FilterArgumentError
from liquid.expression import EMPTY

from liquid.builtin.filters.misc import Size, Default, Date


class Case(NamedTuple):
    description: str
    val: Any
    args: List[Any]
    kwargs: Dict[Any, Any]
    expect: Any


class MiscFilterTestCase(unittest.TestCase):
    """Test miscellaneous filter functions."""

    def _test(self, filter_cls, test_cases):
        """Helper method for running lists of `Case`s"""
        env = Environment()
        func = filter_cls(env)

        for case in test_cases:
            with self.subTest(msg=case.description):
                if isclass(case.expect) and issubclass(
                    case.expect, FilterArgumentError
                ):
                    with self.assertRaises(case.expect):
                        func(case.val, *case.args, **case.kwargs)
                else:
                    self.assertEqual(
                        func(case.val, *case.args, **case.kwargs), case.expect
                    )

    def test_size(self):
        """Test `size` filter function."""
        test_cases = [
            Case(
                description="array", val=["a", "b", "c"], args=[], kwargs={}, expect=3
            ),
            Case(description="string", val="abc", args=[], kwargs={}, expect=3),
            Case(description="empty array", val=[], args=[], kwargs={}, expect=0),
            Case(
                description="unexpected argument",
                val=[1, 2, 3],
                args=["foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not an array or string",
                val=1,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Size, test_cases)

    def test_default(self):
        """Test `default` filter function."""
        test_cases = [
            Case(description="nil", val=None, args=["foo"], kwargs={}, expect="foo"),
            Case(description="false", val=False, args=["foo"], kwargs={}, expect="foo"),
            Case(
                description="empty string",
                val="",
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="empty list",
                val=[],
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="empty object",
                val={},
                args=["foo"],
                kwargs={},
                expect="foo",
            ),
            Case(
                description="not empty string",
                val="hello",
                args=["foo"],
                kwargs={},
                expect="hello",
            ),
            Case(
                description="not empty list",
                val=["hello"],
                args=["foo"],
                kwargs={},
                expect=["hello"],
            ),
            Case(
                description="not empty object",
                val={"title": "hello"},
                args=["foo"],
                kwargs={},
                expect={"title": "hello"},
            ),
            Case(
                description="too many arguments",
                val=None,
                args=["foo", "bar"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing argument",
                val=None,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not an array or string",
                val={},
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="false returns default",
                val=False,
                args=["bar"],
                kwargs={},
                expect="bar",
            ),
            Case(
                description="empty returns default",
                val=EMPTY,
                args=["bar"],
                kwargs={},
                expect="bar",
            ),
            Case(
                description="allow false",
                val=False,
                args=["bar"],
                kwargs={"allow_false": True},
                expect=False,
            ),
        ]

        self._test(Default, test_cases)

    def test_date(self):
        """Test `date` filter function."""
        test_cases = [
            Case(
                description="datetime object",
                val=datetime.datetime(2002, 1, 1, 11, 45, 13),
                args=[r"%a, %b %d, %y"],
                kwargs={},
                expect="Tue, Jan 01, 02",
            ),
            Case(
                description="well formed string",
                val="March 14, 2016",
                args=[r"%b %d, %y"],
                kwargs={},
                expect="Mar 14, 16",
            ),
            Case(
                description="too many arguments",
                val=datetime.datetime(2001, 2, 1, 11, 45, 13),
                args=[r"%a, %b %d, %y", "foo"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="not a date or time",
                val=[],
                args=[r"%a, %b %d, %y"],
                kwargs={},
                expect=FilterArgumentError,
            ),
            Case(
                description="missing argument",
                val=None,
                args=[],
                kwargs={},
                expect=FilterArgumentError,
            ),
        ]

        self._test(Date, test_cases)



"""Test case for the `Template` API."""

import unittest

from liquid import Template
from liquid import Environment
from liquid import Mode


class TemplateAPITestCase(unittest.TestCase):
    """Test case for the `Template` API."""

    def test_implicit_environment(self):
        """Test that an Environment is created automatically."""
        template = Template(r"Hello, {{ you }}")
        self.assertIsNotNone(template.env)

    def test_environment_cache(self):
        """Test that we reuse Environments."""
        template = Template(r"Hello, {{ you }}!")
        another = Template(r"Goodbye, {{ you }}.")

        self.assertEqual(template.env, another.env)

        lax = Template("something", tolerance=Mode.LAX)
        self.assertNotEqual(lax.env, template.env)

    def test_implicit_explicit(self):
        """Test that an implicit environment renders the same as an explicit one."""
        env = Environment()

        source = r"Hello, {{ you }}"
        context = {"you": "there"}

        some = env.from_string(source)
        other = Template(source)

        self.assertEqual(some.render(**context), other.render(**context))



"""Tag definition for illegal or unregistered tags."""

from liquid.ast import IllegalNode
from liquid.exceptions import LiquidSyntaxError
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import TOKEN_ILLEGAL
from liquid.token import TOKEN_EXPRESSION


class Illegal(Tag):
    """Tag definition for illegal or unregistered tags."""

    name = TOKEN_ILLEGAL
    block = False

    def parse(self, stream: TokenStream) -> IllegalNode:
        tok = stream.current
        stream.next_token()

        if stream.current.type == TOKEN_EXPRESSION:
            stream.next_token()

        raise LiquidSyntaxError(f"unexpected tag '{tok.value}'", linenum=tok.linenum)



"""Parse tree node and tag definition for the built in "include" tag."""

import sys

from typing import Optional
from typing import Dict
from typing import Any
from typing import TextIO

from liquid.ast import Node
from liquid.builtin.drops import IterableDrop
from liquid.context import Context

from liquid.expression import Expression
from liquid.expression import Identifier

from liquid.exceptions import LiquidSyntaxError
from liquid.lex import tokenize_include_expression
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.parse import expect
from liquid.parse import parse_identifier
from liquid.parse import parse_expression
from liquid.parse import parse_string_or_identifier
from liquid.parse import parse_unchained_identifier

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_WITH
from liquid.token import TOKEN_FOR
from liquid.token import TOKEN_AS
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_EOF


TAG_INCLUDE = sys.intern("include")


class IncludeNode(Node):
    """Parse tree node for the built-in "include" tag."""

    __slots__ = ("tok", "name", "var", "alias", "args")

    def __init__(
        self,
        tok: Token,
        name: Expression,
        var: Optional[Identifier] = None,
        alias: Optional[str] = None,
        args: Optional[Dict[str, Any]] = None,
    ):
        self.tok = tok
        self.name = name
        self.var = var
        self.alias = alias
        self.args = args or {}

    def __str__(self):
        buf = [f"{self.name}"]

        if self.var:
            buf.append(f" with {self.var}")

        if self.alias:
            buf.append(f" as {self.alias}")

        if self.args:
            buf.append(", ")

        args = (f"{key}={val}" for key, val in self.args.items())
        buf.append(", ".join(args))

        return f"include({''.join(buf)})"

    def __repr__(self):
        return f"IncludeNode(tok={self.tok!r}, name={self.name})"

    def render_to_output(self, context: Context, buffer: TextIO):
        name = self.name.evaluate(context)
        template = context.get_template(str(name))

        namespace: Dict[str, object] = {}

        # Add any keyword arguments to the new template context.
        for key, val in self.args.items():
            namespace[key] = val.evaluate(context)

        with context.extend(namespace):
            # Bind a variable to the included template.
            if self.var is not None:
                val = self.var.evaluate(context)
                key = self.alias or template.name.split(".")[0]

                # If the variable is array-like, render the template once for each
                # item in the array.
                #
                # The reference implementation does not seem to distinguish between
                # "for" and "with". Just checks for array-like-ness.
                if isinstance(val, (tuple, list, IterableDrop)):
                    # NOTE: What if an included template with a bound array updates
                    # a keyword argument value? Do we need to update namespace
                    # arguments after each loop?
                    #
                    # The reference implementation seems to evaluate arguments once,
                    # before the loop.
                    for itm in val:
                        namespace[key] = itm
                        template.render_with_context(context, buffer, partial=True)
                else:
                    namespace[key] = val
                    template.render_with_context(context, buffer, partial=True)
            else:
                template.render_with_context(context, buffer, partial=True)


class IncludeTag(Tag):
    """The built-in "include" tag."""

    name = TAG_INCLUDE
    block = False

    def parse(self, stream: TokenStream) -> IncludeNode:
        """Read an IncludeNode from the given stream of tokens."""
        expect(stream, TOKEN_TAG, value=self.name)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr_stream = TokenStream(tokenize_include_expression(stream.current.value))

        # Need a string or identifier that resolves to a string. This is the name
        # of the template to be included.
        name = parse_string_or_identifier(expr_stream, linenum=tok.linenum)
        expr_stream.next_token()

        # Optionally bind a variable to the included template context
        if expr_stream.current.type in (TOKEN_WITH, TOKEN_FOR):
            expr_stream.next_token()  # Eat 'with' or 'for'
            expect(expr_stream, TOKEN_IDENTIFIER)
            identifier = parse_identifier(expr_stream)
            expr_stream.next_token()

            # The bound variable will take the name of the template by default,
            # or an alias if an identifier follows the "as" keyword.
            if expr_stream.current.type == TOKEN_AS:
                expr_stream.next_token()  # Eat 'as'
                expect(expr_stream, TOKEN_IDENTIFIER)
                alias = str(parse_unchained_identifier(expr_stream))
                expr_stream.next_token()
            else:
                alias = None
        else:
            identifier = None
            alias = None

        # Zero or more keyword arguments
        args = {}

        # The first keyword argument might follow immediately or after a comma.
        if expr_stream.current.type == TOKEN_IDENTIFIER:
            key, val = _parse_argument(expr_stream)
            args[key] = val

        while expr_stream.current.type != TOKEN_EOF:
            if expr_stream.current.type == TOKEN_COMMA:
                expr_stream.next_token()  # Eat comma
                key, val = _parse_argument(expr_stream)
                args[key] = val
            else:
                typ = expr_stream.current.type
                raise LiquidSyntaxError(
                    f"expected a comma separated list of arguments, found {typ}",
                    linenum=tok.linenum,
                )

        return IncludeNode(tok, name=name, var=identifier, alias=alias, args=args)


def _parse_argument(stream: TokenStream):
    key = str(parse_unchained_identifier(stream))
    stream.next_token()

    expect(stream, TOKEN_COLON)
    stream.next_token()  # Eat colon

    val = parse_expression(stream)
    stream.next_token()

    return key, val



"""Tag and node definition for the built-in "unless" tag."""
from __future__ import annotations

import sys

from typing import TextIO
from typing import TYPE_CHECKING

from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.context import Context
from liquid.lex import tokenize_boolean_expression
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.parse import get_parser
from liquid.parse import expect
from liquid.parse import parse_boolean_expression

from liquid.token import Token
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_TAG

if TYPE_CHECKING:
    from liquid import Environment
    from liquid.expression import Expression

TAG_UNLESS = sys.intern("unless")
TAG_ENDUNLESS = sys.intern("endunless")

ENDUNLESSBLOCK = (TAG_ENDUNLESS, TOKEN_EOF)


class UnlessNode(Node):
    """Parse tree node for the built-in "unless" tag."""

    __slots__ = ("tok", "condition", "consequence")

    def __init__(self, tok: Token, condition: Expression, consequence: BlockNode):
        self.tok = tok
        self.condition = condition
        self.consequence = consequence

    def __str__(self) -> str:
        return f"if !{self.condition} {{ {self.consequence} }}"

    def render_to_output(self, context: Context, buffer: TextIO):
        if not self.condition.evaluate(context):
            self.consequence.render(context, buffer)


class UnlessTag(Tag):
    """The built-in "unless" tag."""

    name = TAG_UNLESS
    end = TAG_ENDUNLESS

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse_expression(self, stream: TokenStream) -> Expression:
        expect(stream, TOKEN_EXPRESSION)
        expr_iter = tokenize_boolean_expression(stream.current.value)
        return parse_boolean_expression(TokenStream(expr_iter))

    def parse(self, stream: TokenStream) -> UnlessNode:

        expect(stream, TOKEN_TAG, value=TAG_UNLESS)
        tok = stream.current
        stream.next_token()

        expr = self.parse_expression(stream)
        stream.next_token()

        consequence = self.parser.parse_block(stream, ENDUNLESSBLOCK)

        expect(stream, TOKEN_TAG, value=TAG_ENDUNLESS)
        return UnlessNode(tok, condition=expr, consequence=consequence)



def truncate_chars(val: str, num: int, end: str = "..."):
    """Truncates a string if it is longer than the specified number of characters.
    Truncated strings will end with `end`, an ellipsis by default."""
    val_length = len(val)
    end_length = len(end)

    if val_length < num:
        return val

    return f"{val[:num-end_length]}{end}"


def truncate_words(val: str, num: int, end: str = "..."):
    """Truncates a string down to the number of words passed as the first parameter."""
    # XXX: Replaces consecutive whitespace with a single newline.
    words = val.split()

    if len(words) < num:
        return " ".join(words)

    return " ".join(words[:num]) + end



"""Parse tree node and pseudo "tag" for output statements."""
from typing import TextIO


from liquid.ast import Node
from liquid.expression import Expression
from liquid.context import Context
from liquid.lex import tokenize_filtered_expression
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.parse import parse_filtered_expression
from liquid.parse import expect

from liquid.token import Token
from liquid.token import TOKEN_STATEMENT


class StatementNode(Node):
    """Parse tree node representing an output statement."""

    __slots__ = ("tok", "expression")

    def __init__(self, tok: Token, expression: Expression):
        self.tok = tok
        self.expression = expression

    def __str__(self):
        return f"`{self.expression}`"

    def __repr__(self):  # pragma: no cover
        return f"StatementNode(tok={self.tok}, expression={self.expression!r})"

    def render_to_output(self, context: Context, buffer: TextIO):
        val = self.expression.evaluate(context)
        if isinstance(val, str):
            # shortcut for common case.
            pass
        elif isinstance(val, bool):
            val = str(val).lower()
        elif val is None:
            val = ""
        elif isinstance(val, (list, tuple)):
            val = "".join(str(v) for v in val)
        else:
            val = str(val)

        buffer.write(val)


class Statement(Tag):
    """Pseudo "tag" to register output statements with the environment."""

    name = TOKEN_STATEMENT

    def parse(self, stream: TokenStream) -> Node:
        tok = stream.current
        expect(stream, TOKEN_STATEMENT)

        expr_iter = tokenize_filtered_expression(tok.value)
        node = StatementNode(tok, parse_filtered_expression(TokenStream(expr_iter)))
        return node



"""Tag and node definition for the built-in "decrement" tag."""

import sys
from typing import TextIO

from liquid import ast
from liquid.context import Context
from liquid.lex import tokenize_identifier
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.parse import expect
from liquid.parse import parse_unchained_identifier

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION

TAG_DECREMENT = sys.intern("decrement")


class DecrementNode(ast.Node):
    """Parse tree node for the built-in "decrement" tag."""

    __slots__ = ("tok", "identifier")

    def __init__(self, tok: Token, identifier: str):
        self.tok = tok
        self.identifier = identifier

    def __str__(self) -> str:
        return f"{self.identifier} -= 1"

    def render_to_output(self, context: Context, buffer: TextIO):
        buffer.write(str(context.decrement(self.identifier)))


class DecrementTag(Tag):
    """The built-in "decrement" tag."""

    name = TAG_DECREMENT
    block = False

    def parse(self, stream: TokenStream) -> DecrementNode:
        expect(stream, TOKEN_TAG, value=TAG_DECREMENT)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        tokens = TokenStream(tokenize_identifier(stream.current.value))
        ident = parse_unchained_identifier(tokens)

        return DecrementNode(tok=tok, identifier=str(ident))



"""Liquid template definition."""

from __future__ import annotations

from collections import ChainMap
from collections import abc

from io import StringIO
from pathlib import Path

from typing import Callable
from typing import Dict
from typing import Any
from typing import Optional
from typing import TextIO
from typing import Mapping
from typing import TYPE_CHECKING

from liquid.context import Context

from liquid.exceptions import LiquidInterrupt
from liquid.exceptions import LiquidSyntaxError
from liquid.exceptions import Error


if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment
    from liquid.ast import ParseTree


# pylint: disable=too-many-instance-attributes
class BoundTemplate:
    """A liquid template that has been parsed and is bound to an Environment.

    Rather than instantiating a `BoundTemplate` directly, you'll usually want
    to use the `Template(source)` API or the `get_template` or `from_string`
    methods from an `Environment`.
    """

    # pylint: disable=redefined-builtin
    def __init__(
        self,
        env: Environment,
        parse_tree: ParseTree,
        name: str = "",
        path: Optional[Path] = None,
        globals: Optional[Mapping[str, Any]] = None,
        uptodate: Optional[Callable[[], bool]] = None,
    ):
        self.env = env
        self.tree = parse_tree
        self._globals = globals or {}
        self.name = name
        self.path = path
        self.uptodate = uptodate

    def render(self, *args: ..., **kwargs: ...) -> str:
        """Render the template with `args` and `kwargs` included in the render context.

        Accepts the same arguments as the `dict` constructor.
        """
        _vars: Dict[str, object] = dict(*args, **kwargs)

        context = Context(self.env, ChainMap(_vars, self._globals))

        buf = StringIO()
        self.render_with_context(context, buf)
        return buf.getvalue()

    def render_with_context(
        self,
        context: Context,
        buffer: TextIO,
        *args: ...,
        partial: bool = False,
        block_scope: bool = False,
        **kwargs: ...,
    ):
        """Render the template using an existing context and output buffer.

        Args:
            context: A render context.
            buffer: File-like object to which rendered text is written.
            partial: If `True`, indicates that the current template has been
                included using either a "render" or "include" tag.
            block_scope: If `True`, indicates that assigns, breaks and continues
                from this template will not leak into the parent context.
        """
        # "template" could get overridden from args/kwargs, "partial" will not.
        namespace = self._make_globals(partial, args, kwargs)

        with context.extend(namespace=namespace):
            for node in self.tree.statements:
                try:
                    node.render(context, buffer)
                except LiquidInterrupt as err:
                    # If this is an "included" template, there could be a for loop
                    # in a parent template. A for loop that could be interrupted
                    # from an included template.
                    #
                    # Convert the interrupt to a syntax error if there is no parent.
                    if not partial or block_scope:
                        self.env.error(
                            LiquidSyntaxError(
                                f"unexpected '{err}'", linenum=node.tok.linenum
                            )
                        )
                    else:
                        raise
                except Error as err:
                    # Raise or warn according to the current mode.
                    self.env.error(err, linenum=node.tok.linenum)

    @property
    def is_up_to_date(self) -> bool:
        """False if the template was modified since it was last parsed,
        True otherwise."""
        if not self.uptodate:
            return True
        return self.uptodate()

    def _make_globals(self, partial, args, kwargs) -> abc.Mapping:
        return {
            **dict(*args, **kwargs),
            "partial": partial,
        }

    def __repr__(self):
        return (
            f"Template(name='{self.name}', "
            f"path='{self.path}', uptodate={self.is_up_to_date})"
        )  # pragma: no cover


class AwareBoundTemplate(BoundTemplate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.drop = TemplateDrop(self.name, self.path)

    def _make_globals(self, partial, args, kwargs) -> abc.Mapping:
        return {"template": self.drop, **super()._make_globals(partial, args, kwargs)}


class TemplateDrop(abc.Mapping):
    """Template meta data mapping."""

    def __init__(self, name: str, path: Optional[Path]):
        self.name = name
        self.path = path or Path(name)

        self.stem = self.path.stem
        if "." in self.stem:
            self.suffix = self.stem.split(".")[-1]
        else:
            self.suffix = None

        self._items = {
            "directory": self.path.parent.name,
            "name": self.path.name.split(".")[0],
            "suffix": self.suffix,
        }

    def __str__(self):
        return self.stem

    def __repr__(self):
        return (
            f"TemplateDrop(directory='{self['directory']}', "
            f"name='{self['name']}', suffix='{self['suffix']}')"
        )  # pragma: no cover

    def __contains__(self, item):
        return item in self._items

    def __getitem__(self, key):
        return self._items[key]

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return iter(self._items)



"""Tag and node definition for the built-in "liquid" tag."""

import sys
from typing import TextIO

from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.context import Context
from liquid.lex import tokenize_liquid_expression

from liquid.parse import expect
from liquid.parse import get_parser

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION


TAG_LIQUID = sys.intern("liquid")


class LiquidNode(Node):
    """Parse tree node for the built-in "liquid" tag."""

    __slots__ = ("tok", "block")

    def __init__(self, tok: Token, block: BlockNode):
        self.tok = tok
        self.block = block

    def __str__(self):
        return str(self.block)

    def __repr__(self):  # pragma: no cover
        return f"LiquidNode(tok={self.tok}, block={self.block!r})"

    def render_to_output(self, context: Context, buffer: TextIO):
        return self.block.render(context, buffer)


class LiquidTag(Tag):
    """The built-in "liquid" tag."""

    name = TAG_LIQUID
    block = False

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_LIQUID)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr_stream = TokenStream(tokenize_liquid_expression(stream.current.value))

        parser = get_parser(self.env)
        block = parser.parse_block(expr_stream, end=())
        return LiquidNode(tok, block=block)



"""Tag and node definition for the built-in "ifchanged" tag."""
from __future__ import annotations
import sys

from io import StringIO

from typing import TextIO
from typing import TYPE_CHECKING

from liquid.token import Token
from liquid.token import TOKEN_TAG

from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.parse import expect
from liquid.parse import get_parser

from liquid.stream import TokenStream
from liquid.tag import Tag

if TYPE_CHECKING:  # pragma: no cover
    from liquid.context import Context
    from liquid import Environment


TAG_IFCHANGED = sys.intern("ifchanged")
TAG_ENDIFCHANGED = sys.intern("endifchanged")

ENDIFCHANGEDBLOCK = (TAG_ENDIFCHANGED,)


class IfChangedNode(Node):
    """Parse tree node for the built-in "ifchanged" tag"""

    __slots__ = ("tok", "block")

    def __init__(self, tok: Token, block: BlockNode):
        self.tok = tok
        self.block = block

    def __str__(self) -> str:
        return f"ifchanged {{ {self.block} }}"

    def __repr__(self):  # pragma: no cover
        return f"IfChanged(tok={self.tok})"

    def render_to_output(self, context: Context, buffer: TextIO):
        # Render to an intermediate buffer.
        buf = StringIO()
        self.block.render(context, buf)
        val = buf.getvalue()

        # The context will update its namespace if needed.
        if context.ifchanged(val):
            buffer.write(val)


class IfChangedTag(Tag):
    """The built-in "ifchanged" tag"""

    name = TAG_IFCHANGED
    end = TAG_ENDIFCHANGED

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_IFCHANGED)
        tok = stream.current
        stream.next_token()

        block = self.parser.parse_block(stream, ENDIFCHANGEDBLOCK)
        expect(stream, TOKEN_TAG, value=TAG_ENDIFCHANGED)

        return IfChangedNode(tok=tok, block=block)



"""Parse tree node and tag definition for the built in "render" tag."""

import pathlib
import sys

from typing import Optional
from typing import Dict
from typing import Any
from typing import TextIO

from liquid.ast import Node

from liquid.builtin.drops import IterableDrop
from liquid.builtin.tags.for_tag import ForLoop
from liquid.builtin.tags.include_tag import TAG_INCLUDE

from liquid.context import Context
from liquid.context import ReadOnlyChainMap

from liquid.exceptions import LiquidSyntaxError
from liquid.expression import Expression
from liquid.lex import tokenize_include_expression

from liquid.parse import expect
from liquid.parse import parse_identifier
from liquid.parse import parse_expression
from liquid.parse import parse_unchained_identifier
from liquid.parse import parse_string_literal

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_WITH
from liquid.token import TOKEN_FOR
from liquid.token import TOKEN_AS
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_STRING


TAG_RENDER = sys.intern("render")


class RenderNode(Node):
    """Parse tree node for the built-in "render" tag."""

    __slots__ = ("tok", "name", "var", "loop", "alias", "args")

    def __init__(
        self,
        tok: Token,
        name: Expression,
        var: Optional[Expression] = None,
        loop: bool = False,
        alias: Optional[str] = None,
        args: Optional[Dict[str, Any]] = None,
    ):
        self.tok = tok
        self.name = name
        self.var = var
        self.loop = loop
        self.alias = alias
        self.args = args

    def __str__(self):
        buf = [f"{self.name}"]

        if self.var:
            buf.append(f" with {self.var}")

        if self.alias:
            buf.append(f" as {self.alias}")

        if self.args:
            buf.append(", ")

        args = (f"{key}={val}" for key, val in self.args.items())
        buf.append(", ".join(args))

        return f"render({''.join(buf)})"

    def __repr__(self):
        return f"RenderNode(tok={self.tok!r}, name={self.name})"

    def render_to_output(self, context: Context, buffer: TextIO):
        template_name = self.name.evaluate(context)

        # TODO: Store this kind of thing on the context object but not the
        # globals/locals namespace.
        render_folder = context.get("render_folder", default=None)
        if render_folder:
            path = pathlib.Path(str(render_folder), str(template_name)).with_suffix(
                ".liquid"
            )
        else:
            path = template_name

        template = context.get_template(str(path))

        # Evaluate keyword arguments once. Unlike 'include', 'render' can not
        # mutate variables in the outer scope, so there's no need to re-evaluate
        # arguments for each loop (if any).
        args = {k: v.evaluate(context) for k, v in self.args.items()}

        # We're using a chain map here in case we need to push a forloop drop into
        # it. As drops are read only, the built-in collections.ChainMap will not do.
        namespace = ReadOnlyChainMap(args)

        # New context with globals and filters from the parent, plus the read only
        # namespace containing render arguments and bound variable.
        ctx = context.copy(namespace, disabled_tags=[TAG_INCLUDE])

        # Optionally bind a variable to the render namespace.
        if self.var is not None:
            val = self.var.evaluate(context)
            key = self.alias or template.name.split(".")[0]

            # If the variable is array-like, render the template once for each item.
            # `self.loop` being True indicates the render expression used "for" not
            # "with". This distinction is not made when using the 'include' tag.
            if self.loop and isinstance(val, (tuple, list, IterableDrop)):
                forloop = ForLoop(
                    name=key,
                    it=iter(val),
                    length=len(val),
                )

                args["forloop"] = forloop
                args[key] = None

                for itm in forloop:
                    args[key] = itm
                    template.render_with_context(
                        ctx, buffer, partial=True, block_scope=True
                    )
            else:
                # The bound variable is not array-like, shove it into the namespace
                # via args.
                args[key] = val
                template.render_with_context(
                    ctx, buffer, partial=True, block_scope=True
                )
        else:
            template.render_with_context(ctx, buffer, partial=True, block_scope=True)


class RenderTag(Tag):
    """The built-in "render" tag."""

    name = TAG_RENDER
    block = False

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_RENDER)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr_stream = TokenStream(tokenize_include_expression(stream.current.value))

        # Need a string. 'render' does not accept identifiers that resolve to a string.
        # This is the name of the template to be included.
        expect(expr_stream, TOKEN_STRING)
        name = parse_string_literal(expr_stream)
        expr_stream.next_token()

        # Optionally bind a variable to the included template context
        if expr_stream.current.type in (TOKEN_WITH, TOKEN_FOR):
            loop = expr_stream.current.type == TOKEN_FOR
            expr_stream.next_token()  # Eat 'with' or 'for'
            expect(expr_stream, TOKEN_IDENTIFIER)
            identifier = parse_identifier(expr_stream)
            expr_stream.next_token()

            # The bound variable will take the name of the template by default,
            # or an alias if an identifier follows the "as" keyword.
            if expr_stream.current.type == TOKEN_AS:
                expr_stream.next_token()  # Eat 'as'
                expect(expr_stream, TOKEN_IDENTIFIER)
                alias = str(parse_unchained_identifier(expr_stream))
                expr_stream.next_token()
            else:
                alias = None
        else:
            identifier = None
            loop = False
            alias = None

        # Zero or more keyword arguments
        args = {}

        # The first keyword argument might follow immediately or after a comma.
        if expr_stream.current.type == TOKEN_IDENTIFIER:
            key, val = parse_argument(expr_stream)
            args[key] = val

        while expr_stream.current.type != TOKEN_EOF:
            if expr_stream.current.type == TOKEN_COMMA:
                expr_stream.next_token()  # Eat comma
                key, val = parse_argument(expr_stream)
                args[key] = val
            else:
                typ = expr_stream.current.type
                raise LiquidSyntaxError(
                    f"expected a comma separated list of arguments, found {typ}",
                    linenum=tok.linenum,
                )

        return RenderNode(
            tok,
            name=name,
            var=identifier,
            loop=loop,
            alias=alias,
            args=args,
        )


def parse_argument(stream: TokenStream):
    """Return the next key/value pair from the stream where key and value
    are separated by a colon."""
    key = str(parse_unchained_identifier(stream))
    stream.next_token()

    expect(stream, TOKEN_COLON)
    stream.next_token()  # Eat colon

    val = parse_expression(stream)
    stream.next_token()

    return key, val



"""Tag and node definition for the built-in "case" tag."""
from __future__ import annotations

import sys

from typing import Optional
from typing import List
from typing import TextIO
from typing import TYPE_CHECKING

from liquid.token import Token
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_TAG

from liquid.parse import get_parser
from liquid.parse import expect
from liquid.parse import parse_boolean_expression

from liquid import ast
from liquid.tag import Tag
from liquid.context import Context
from liquid.stream import TokenStream
from liquid.lex import tokenize_boolean_expression

if TYPE_CHECKING:
    from liquid import Environment
    from liquid.expression import Expression

TAG_CASE = sys.intern("case")
TAG_ENDCASE = sys.intern("endcase")
TAG_WHEN = sys.intern("when")
TAG_ELSE = sys.intern("else")

ENDWHENBLOCK = (TAG_ENDCASE, TAG_WHEN, TAG_ELSE, TOKEN_EOF)
ENDCASEBLOCK = (TAG_ENDCASE,)


class CaseNode(ast.Node):
    """Parse tree node for the built-in "case" tag."""

    __slots__ = ("tok", "whens", "default")

    def __init__(
        self,
        tok: Token,
        whens: List[ast.ConditionalBlockNode],
        default: Optional[ast.BlockNode] = None,
    ):
        self.tok = tok
        self.whens = whens
        self.default = default

    def __str__(self) -> str:
        if not self.whens:
            buf = ["if (False) { }"]
        else:
            buf = [f"if {self.whens[0]}"]

        if len(self.whens) > 1:
            for when in self.whens[1:]:
                buf.append(f"elsif {when}")

        if self.default:
            buf.append(f"else {{ {self.default} }}")

        return " ".join(buf)

    def render_to_output(self, context: Context, buffer: TextIO):
        rendered = False

        for when in self.whens:
            if when.render(context, buffer):
                rendered = True
                break

        if not rendered and self.default:
            self.default.render(context, buffer)


class CaseTag(Tag):
    """The built-in cycle tag."""

    name = TAG_CASE
    end = TAG_ENDCASE

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse_expression(self, case: str, stream: TokenStream) -> Expression:
        expect(stream, TOKEN_EXPRESSION)
        expr_iter = tokenize_boolean_expression(f"{case} == {stream.current.value}")
        return parse_boolean_expression(TokenStream(expr_iter))

    def parse(self, stream: TokenStream) -> ast.Node:
        expect(stream, TOKEN_TAG, value=TAG_CASE)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        case = stream.current.value
        stream.next_token()

        # Eat whitespace or junk between `case` and when/else/endcase
        while (
            stream.current.type != TOKEN_TAG
            and stream.current.value not in ENDWHENBLOCK
        ):
            stream.next_token()

        whens = []

        while stream.current.istag(TAG_WHEN):
            stream.next_token()  # Eat WHEN
            expr = self.parse_expression(case, stream)

            when = ast.ConditionalBlockNode(
                stream.current,
                condition=expr,
            )

            stream.next_token()
            when.block = self.parser.parse_block(stream, ENDWHENBLOCK)
            whens.append(when)

        if stream.current.istag(TAG_ELSE):
            stream.next_token()
            default = self.parser.parse_block(stream, ENDCASEBLOCK)
        else:
            default = None

        expect(stream, TOKEN_TAG, value=TAG_ENDCASE)
        return CaseNode(tok, whens=whens, default=default)



__version__ = "0.7.1"

from .mode import Mode
from .filter import Filter
from .token import Token
from .context import Context
from .expression import Expression

from .loaders import FileSystemLoader

from .environment import Environment
from .environment import Template



from __future__ import annotations

from typing import TYPE_CHECKING

from . import literal
from . import statement
from . import illegal

from .tags import if_tag
from .tags import comment_tag
from .tags import unless_tag
from .tags import case_tag
from .tags import for_tag
from .tags import tablerow_tag
from .tags import capture_tag
from .tags import cycle_tag
from .tags import assign_tag
from .tags import increment_tag
from .tags import decrement_tag
from .tags import echo_tag
from .tags import liquid_tag
from .tags import include_tag
from .tags import render_tag
from .tags import ifchanged_tag


from .filters.math import (
    Abs,
    AtMost,
    AtLeast,
    Ceil,
    DividedBy,
    Floor,
    Minus,
    Plus,
    Round,
    Times,
    Modulo,
)

from .filters.string import (
    Capitalize,
    Append,
    Downcase,
    Escape,
    EscapeOnce,
    LStrip,
    NewlineToBR,
    Prepend,
    Remove,
    RemoveFirst,
    Replace,
    ReplaceFirst,
    Slice,
    Split,
    Upcase,
    Strip,
    RStrip,
    StripHTML,
    StripNewlines,
    Truncate,
    TruncateWords,
    URLEncode,
    URLDecode,
)

from .filters.array import (
    Join,
    First,
    Last,
    Concat,
    Map,
    Reverse,
    Sort,
    SortNatural,
    Where,
    Uniq,
    Compact,
)

from .filters.misc import Size, Default, Date


if TYPE_CHECKING:
    from liquid import Environment


def register(env: Environment) -> None:
    env.add_tag(literal.Literal)
    env.add_tag(statement.Statement)
    env.add_tag(illegal.Illegal)
    env.add_tag(if_tag.IfTag)
    env.add_tag(comment_tag.CommentTag)
    env.add_tag(unless_tag.UnlessTag)
    env.add_tag(case_tag.CaseTag)
    env.add_tag(for_tag.ForTag)
    env.add_tag(for_tag.BreakTag)
    env.add_tag(for_tag.ContinueTag)
    env.add_tag(tablerow_tag.TablerowTag)
    env.add_tag(capture_tag.CaptureTag)
    env.add_tag(cycle_tag.CycleTag)
    env.add_tag(assign_tag.AssignTag)
    env.add_tag(increment_tag.IncrementTag)
    env.add_tag(decrement_tag.DecrementTag)
    env.add_tag(echo_tag.EchoTag)
    env.add_tag(liquid_tag.LiquidTag)
    env.add_tag(include_tag.IncludeTag)
    env.add_tag(render_tag.RenderTag)
    env.add_tag(ifchanged_tag.IfChangedTag)

    env.add_filter(Abs.name, Abs(env))
    env.add_filter(AtMost.name, AtMost(env))
    env.add_filter(AtLeast.name, AtLeast(env))
    env.add_filter(Ceil.name, Ceil(env))
    env.add_filter(DividedBy.name, DividedBy(env))
    env.add_filter(Floor.name, Floor(env))
    env.add_filter(Minus.name, Minus(env))
    env.add_filter(Plus.name, Plus(env))
    env.add_filter(Round.name, Round(env))
    env.add_filter(Times.name, Times(env))
    env.add_filter(Modulo.name, Modulo(env))

    env.add_filter(Capitalize.name, Capitalize(env))
    env.add_filter(Append.name, Append(env))
    env.add_filter(Downcase.name, Downcase(env))
    env.add_filter(Escape.name, Escape(env))
    env.add_filter(EscapeOnce.name, EscapeOnce(env))
    env.add_filter(LStrip.name, LStrip(env))
    env.add_filter(NewlineToBR.name, NewlineToBR(env))
    env.add_filter(Prepend.name, Prepend(env))
    env.add_filter(Remove.name, Remove(env))
    env.add_filter(RemoveFirst.name, RemoveFirst(env))
    env.add_filter(Replace.name, Replace(env))
    env.add_filter(ReplaceFirst.name, ReplaceFirst(env))
    env.add_filter(Slice.name, Slice(env))
    env.add_filter(Split.name, Split(env))
    env.add_filter(Upcase.name, Upcase(env))
    env.add_filter(Strip.name, Strip(env))
    env.add_filter(RStrip.name, RStrip(env))
    env.add_filter(StripHTML.name, StripHTML(env))
    env.add_filter(StripNewlines.name, StripNewlines(env))
    env.add_filter(Truncate.name, Truncate(env))
    env.add_filter(TruncateWords.name, TruncateWords(env))
    env.add_filter(URLEncode.name, URLEncode(env))
    env.add_filter(URLDecode.name, URLDecode(env))

    env.add_filter(Join.name, Join(env))
    env.add_filter(First.name, First(env))
    env.add_filter(Last.name, Last(env))
    env.add_filter(Concat.name, Concat(env))
    env.add_filter(Map.name, Map(env))
    env.add_filter(Reverse.name, Reverse(env))
    env.add_filter(Sort.name, Sort(env))
    env.add_filter(SortNatural.name, SortNatural(env))
    env.add_filter(Where.name, Where(env))
    env.add_filter(Uniq.name, Uniq(env))
    env.add_filter(Compact.name, Compact(env))

    env.add_filter(Size.name, Size(env))
    env.add_filter(Default.name, Default(env))
    env.add_filter(Date.name, Date(env))



"""Tag and node definition for the built-in "echo" tag."""

import sys


from liquid.ast import Node
from liquid.builtin.statement import StatementNode
from liquid.lex import tokenize_filtered_expression

from liquid.parse import parse_filtered_expression
from liquid.parse import expect

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION

TAG_ECHO = sys.intern("echo")


class EchoNode(StatementNode):
    """Parse tree node for the built-in "echo" tag."""

    __slots__ = ("tok", "expression")

    def __repr__(self):  # pragma: no cover
        return f"EchoNode(tok={self.tok}, expression={self.expression!r})"


class EchoTag(Tag):
    """The built-in "echo" tag."""

    name = TAG_ECHO
    block = False

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_ECHO)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr_iter = tokenize_filtered_expression(stream.current.value)

        expr = parse_filtered_expression(TokenStream(expr_iter))
        return EchoNode(tok, expression=expr)



"""Bad context test cases."""

from unittest import TestCase
from typing import NamedTuple, Type

from liquid.environment import Environment
from liquid.mode import Mode
from liquid.exceptions import LiquidTypeError, lookup_warning


class Case(NamedTuple):
    """Table driven test case helper."""

    description: str
    template: str
    expect_exception: Type[Exception]
    expect_msg: str
    expect_render: str = ""


class BadContextTemplateTestCase(TestCase):
    """Bad context test case."""

    def _test(self, test_cases, mode=Mode.STRICT):
        """Helper method for running lists of `Case`s"""
        env = Environment()
        env.mode = mode

        global_context = {"arr": [], "hash": {}}

        for case in test_cases:
            with self.subTest(msg=case.description):
                if mode == Mode.STRICT:
                    with self.assertRaises(case.expect_exception) as raised:
                        template = env.from_string(
                            case.template, globals=global_context
                        )
                        result = template.render()

                    self.assertEqual(str(raised.exception), case.expect_msg)

                elif mode == Mode.WARN:
                    with self.assertWarns(lookup_warning(case.expect_exception)):
                        template = env.from_string(
                            case.template, globals=global_context
                        )
                        result = template.render()

                elif mode == Mode.LAX:
                    template = env.from_string(case.template, globals=global_context)
                    result = template.render()
                    self.assertEqual(result, case.expect_render)

    def test_bad_context(self):
        """Test that we gracefully handle render time errors due to incorrect context."""
        test_cases = [
            Case(
                description="loop expression variable does not exist",
                template="{% for tag in product.tags %}{tag}{% endfor %}",
                expect_exception=LiquidTypeError,
                expect_msg="expected array or hash at 'product.tags', found 'None', on line 1",
            ),
            Case(
                description="array less than hash",
                template="{% if arr < hash %}foo{% endif %}",
                expect_exception=LiquidTypeError,
                expect_msg=r"invalid operator for types '[] < {}', on line 1",
            ),
        ]

        self._test(test_cases, mode=Mode.STRICT)
        self._test(test_cases, mode=Mode.WARN)
        self._test(test_cases, mode=Mode.LAX)



"""Tag and node definition for the built-in "cycle" tag."""

import sys

from typing import List
from typing import Any
from typing import Optional
from typing import TextIO

from liquid.ast import Node
from liquid.context import Context
from liquid.exceptions import LiquidSyntaxError
from liquid.expression import Expression
from liquid.lex import tokenize_filtered_expression

from liquid.parse import expect
from liquid.parse import parse_expression
from liquid.parse import parse_string_or_identifier

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_COLON

TAG_CYCLE = sys.intern("cycle")


class CycleNode(Node):
    """Parse tree node for the built-in "cycle" tag."""

    __slots__ = ("tok", "group", "args", "key")

    def __init__(self, tok: Token, group: Optional[Expression], args: List[Any]):
        self.tok = tok
        self.group = group
        self.args = args

    def __str__(self) -> str:
        buf = ["cycle ("]
        if self.group:
            buf.append(f"{self.group}: ")

        buf.append(", ".join([str(arg) for arg in self.args]))
        buf.append(")")
        return "".join(buf)

    def render_to_output(self, context: Context, buffer: TextIO):
        if self.group:
            group_name = str(self.group.evaluate(context))
        else:
            group_name = ""

        args = [arg.evaluate(context) for arg in self.args]
        buffer.write(str(next(context.cycle(group_name, args))))


class CycleTag(Tag):
    """The built-in "cycle" tag."""

    name = TAG_CYCLE
    block = False

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_CYCLE)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr_stream = TokenStream(tokenize_filtered_expression(stream.current.value))

        if ":" in stream.current.value:
            group_name = parse_string_or_identifier(expr_stream, linenum=tok.linenum)
            expr_stream.next_token()

            expect(expr_stream, TOKEN_COLON)
            expr_stream.next_token()

        else:
            group_name = None

        args = []
        while expr_stream.current.type != TOKEN_EOF:
            val = parse_expression(expr_stream)
            args.append(val)
            expr_stream.next_token()

            if expr_stream.current.type == TOKEN_COMMA:
                expr_stream.next_token()  # Eat comma
            elif expr_stream.current.type == TOKEN_EOF:
                break
            else:
                raise LiquidSyntaxError(
                    f"expected a comma separated list of arguments, "
                    f"found {expr_stream.current.type}",
                    linenum=tok.linenum,
                )

        return CycleNode(tok, group_name, args)



"""Liquid expression parser test cases."""

import unittest
from typing import NamedTuple, Any

from liquid.expression import FilteredExpression
from liquid.expression import BooleanExpression
from liquid.expression import AssignmentExpression
from liquid.expression import LoopExpression
from liquid.expression import IdentifierPathElement
from liquid.expression import Identifier
from liquid.expression import Boolean
from liquid.expression import Filter
from liquid.expression import StringLiteral
from liquid.expression import IntegerLiteral
from liquid.expression import FloatLiteral
from liquid.expression import PrefixExpression
from liquid.expression import InfixExpression

from liquid.token import Token
from liquid.token import TOKEN_INITIAL

from liquid.stream import TokenStream

from liquid.lex import tokenize_loop_expression
from liquid.lex import tokenize_filtered_expression
from liquid.lex import tokenize_boolean_expression
from liquid.lex import tokenize_assignment_expression

from liquid.parse import parse_filtered_expression
from liquid.parse import parse_boolean_expression
from liquid.parse import parse_assignment_expression
from liquid.parse import parse_loop_expression


class Case(NamedTuple):
    """Subtest helper"""

    description: str
    expression: str
    expect: Any


MockToken = Token(1, TOKEN_INITIAL, "MOCK")


class LiquidFilteredExpressionParserTestCase(unittest.TestCase):
    """Liquid expression parser test cases."""

    def _test(self, test_cases, lex_func, parse_func):
        """Helper method for testing lists of Cases."""

        for case in test_cases:
            with self.subTest(msg=case.description):
                tokens = TokenStream(lex_func(case.expression))
                expr = parse_func(tokens)
                self.assertEqual(expr, case.expect)

    def test_parse_filtered_expression(self):
        """Test that we can parse liquid statement expressions."""
        test_cases = [
            Case(
                "string literal single quotes",
                "'foobar'",
                FilteredExpression(
                    expression=StringLiteral("foobar"),
                    filters=[],
                ),
            ),
            Case(
                "string literal double quotes",
                '"foobar"',
                FilteredExpression(
                    expression=StringLiteral("foobar"),
                    filters=[],
                ),
            ),
            Case(
                "integer literal",
                "7",
                FilteredExpression(
                    expression=IntegerLiteral(7),
                    filters=[],
                ),
            ),
            Case(
                "negative integer literal statement expression",
                "-7",
                FilteredExpression(
                    expression=PrefixExpression(
                        "-",
                        right=IntegerLiteral(7),
                    ),
                    filters=[],
                ),
            ),
            Case(
                "float literal statement expression",
                "3.14",
                FilteredExpression(
                    expression=FloatLiteral(3.14),
                    filters=[],
                ),
            ),
            Case(
                "negative float literal statement expression",
                "-3.14",
                FilteredExpression(
                    expression=PrefixExpression(
                        "-",
                        right=FloatLiteral(3.14),
                    ),
                    filters=[],
                ),
            ),
            Case(
                "single identifier statement expression",
                "collection",
                FilteredExpression(
                    expression=Identifier(
                        path=[IdentifierPathElement("collection")],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier",
                "collection.products",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by double quoted key",
                'collection["products"]',
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by double quoted key followed by index",
                'collection["products"][0]',
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement(0),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by double quoted key followed by identifier",
                'collection["products"].title',
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement("title"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier by single quoted key",
                "collection['products']",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with array index",
                "collection.products[0]",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement(0),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with array index from variable",
                "collection.products[i]",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            Identifier(
                                path=[IdentifierPathElement("i")],
                            ),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with array index from chained identifier",
                "collection.products[some.['object'].count]",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            Identifier(
                                path=[
                                    IdentifierPathElement("some"),
                                    IdentifierPathElement("object"),
                                    IdentifierPathElement("count"),
                                ],
                            ),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with object from chained identifier",
                "linklists[section.settings.menu]",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("linklists"),
                            Identifier(
                                path=[
                                    IdentifierPathElement("section"),
                                    IdentifierPathElement("settings"),
                                    IdentifierPathElement("menu"),
                                ],
                            ),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "chained identifier with object from chained identifier and trailing identifier",
                "linklists[section.settings.menu].links",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("linklists"),
                            Identifier(
                                path=[
                                    IdentifierPathElement("section"),
                                    IdentifierPathElement("settings"),
                                    IdentifierPathElement("menu"),
                                ],
                            ),
                            IdentifierPathElement("links"),
                        ],
                    ),
                    filters=[],
                ),
            ),
            Case(
                "string literal with filter",
                "'foo' | upcase",
                FilteredExpression(
                    expression=StringLiteral("foo"),
                    filters=[Filter(name="upcase", args=[])],
                ),
            ),
            Case(
                "identifier with filter",
                "collection.title | upcase",
                FilteredExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("title"),
                        ],
                    ),
                    filters=[Filter(name="upcase", args=[])],
                ),
            ),
            Case(
                "integer literal with filter and integer argument",
                "4 | at_least: 5",
                FilteredExpression(
                    expression=IntegerLiteral(4),
                    filters=[Filter(name="at_least", args=[IntegerLiteral(5)])],
                ),
            ),
            Case(
                "float literal with filter and float argument",
                "4.1 | divided_by: 5.2",
                FilteredExpression(
                    expression=FloatLiteral(4.1),
                    filters=[Filter(name="divided_by", args=[FloatLiteral(5.2)])],
                ),
            ),
            Case(
                "string literal with filter and string argument",
                "'foo' | append: 'bar'",
                FilteredExpression(
                    expression=StringLiteral("foo"),
                    filters=[Filter(name="append", args=[StringLiteral("bar")])],
                ),
            ),
            Case(
                "string literal with filter and identifier argument",
                "'foo' | append: collection.title",
                FilteredExpression(
                    expression=StringLiteral("foo"),
                    filters=[
                        Filter(
                            name="append",
                            args=[
                                Identifier(
                                    path=[
                                        IdentifierPathElement("collection"),
                                        IdentifierPathElement("title"),
                                    ],
                                )
                            ],
                        )
                    ],
                ),
            ),
            Case(
                "string literal with filter and multiple arguments",
                '"Liquid" | slice: 2, 5',
                FilteredExpression(
                    expression=StringLiteral("Liquid"),
                    filters=[
                        Filter(
                            name="slice",
                            args=[
                                IntegerLiteral(2),
                                IntegerLiteral(5),
                            ],
                        )
                    ],
                ),
            ),
            Case(
                "string literal with multiple filters",
                '"Liquid" | slice: 2, 5 | upcase',
                FilteredExpression(
                    expression=StringLiteral("Liquid"),
                    filters=[
                        Filter(
                            name="slice",
                            args=[
                                IntegerLiteral(2),
                                IntegerLiteral(5),
                            ],
                        ),
                        Filter(name="upcase", args=[]),
                    ],
                ),
            ),
            Case(
                "inconsistent whitespace",
                ' "Liquid"   |slice: 2,5',
                FilteredExpression(
                    expression=StringLiteral("Liquid"),
                    filters=[
                        Filter(
                            name="slice",
                            args=[
                                IntegerLiteral(2),
                                IntegerLiteral(5),
                            ],
                        )
                    ],
                ),
            ),
        ]

        self._test(test_cases, tokenize_filtered_expression, parse_filtered_expression)

    def test_parse_boolean_expression(self):
        """Test that we can parse boolean expressions."""
        test_cases = [
            Case(
                "string literal double quotes",
                '"foobar"',
                BooleanExpression(
                    expression=StringLiteral("foobar"),
                ),
            ),
            Case(
                "integer literal",
                "7",
                BooleanExpression(
                    expression=IntegerLiteral(7),
                ),
            ),
            Case(
                "negative integer literal statement expression",
                "-7",
                BooleanExpression(
                    expression=PrefixExpression(
                        "-",
                        right=IntegerLiteral(7),
                    ),
                ),
            ),
            Case(
                "float literal statement expression",
                "3.14",
                BooleanExpression(
                    expression=FloatLiteral(3.14),
                ),
            ),
            Case(
                "negative float literal statement expression",
                "-3.14",
                BooleanExpression(
                    expression=PrefixExpression(
                        "-",
                        right=FloatLiteral(3.14),
                    ),
                ),
            ),
            Case(
                "single identifier statement expression",
                "collection",
                BooleanExpression(
                    expression=Identifier(
                        path=[IdentifierPathElement("collection")],
                    ),
                ),
            ),
            Case(
                "chained identifier",
                "collection.products",
                BooleanExpression(
                    expression=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                ),
            ),
            Case(
                "keyword true",
                "true",
                BooleanExpression(
                    expression=Boolean(True),
                ),
            ),
            Case(
                "keyword false",
                "false",
                BooleanExpression(
                    expression=Boolean(False),
                ),
            ),
            Case(
                "boolean equality",
                "true == true",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="==",
                        right=Boolean(True),
                    ),
                ),
            ),
            Case(
                "boolean inequality",
                "true != false",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="!=",
                        right=Boolean(False),
                    ),
                ),
            ),
            Case(
                "boolean inequality alternate",
                "true <> false",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="<>",
                        right=Boolean(False),
                    ),
                ),
            ),
            Case(
                "identifier greater than literal",
                "user.age > 21",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[
                                IdentifierPathElement("user"),
                                IdentifierPathElement("age"),
                            ],
                        ),
                        operator=">",
                        right=IntegerLiteral(21),
                    ),
                ),
            ),
            Case(
                "identifier less than literal",
                "age < 18",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[IdentifierPathElement("age")],
                        ),
                        operator="<",
                        right=IntegerLiteral(18),
                    ),
                ),
            ),
            Case(
                "identifier less than or equal to literal",
                "age <= 18",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[IdentifierPathElement("age")],
                        ),
                        operator="<=",
                        right=IntegerLiteral(18),
                    ),
                ),
            ),
            Case(
                "identifier greater than or equal to literal",
                "age >= 18",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[IdentifierPathElement("age")],
                        ),
                        operator=">=",
                        right=IntegerLiteral(18),
                    ),
                ),
            ),
            Case(
                "boolean or boolean",
                "true or false",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Boolean(True),
                        operator="or",
                        right=Boolean(False),
                    ),
                ),
            ),
            Case(
                "identifier contains string",
                "product.tags contains 'sale'",
                BooleanExpression(
                    expression=InfixExpression(
                        left=Identifier(
                            path=[
                                IdentifierPathElement("product"),
                                IdentifierPathElement("tags"),
                            ],
                        ),
                        operator="contains",
                        right=StringLiteral("sale"),
                    ),
                ),
            ),
        ]

        self._test(test_cases, tokenize_boolean_expression, parse_boolean_expression)

    def test_parse_boolean_expression_precedence(self):
        """Test that we get the expected precedence when parsing boolean expressions."""
        test_cases = [
            Case(
                "boolean and logic operators",
                "true and false and false or true",
                "(True and (False and (False or True)))",
            ),
            Case(
                "equality and logic operators",
                "user.name == 'bob' and user.age < 50 or collection.title == 'offers'",
                "((user.name == 'bob') and ((user.age < 50) or (collection.title == 'offers')))",
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                tokens = TokenStream(tokenize_boolean_expression(case.expression))
                expr = parse_boolean_expression(tokens)
                self.assertEqual(str(expr), case.expect)

    def test_parse_assignment_expression(self):
        """Test that we can parse assignment expressions."""
        test_cases = [
            Case(
                "assign a string literal",
                "some = 'foo'",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=StringLiteral(value="foo"),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign an integer literal",
                "some = 5",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=IntegerLiteral(value=5),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign a float literal",
                "some = 5.6",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=FloatLiteral(value=5.6),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign a negative integer literal",
                "some = -5.6",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=PrefixExpression(
                            operator="-",
                            right=FloatLiteral(value=5.6),
                        ),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign a boolean",
                "some = true",
                AssignmentExpression(
                    name="some",
                    expression=FilteredExpression(
                        expression=Boolean(value=True),
                        filters=[],
                    ),
                ),
            ),
            Case(
                "assign with a filter",
                'fruits = "apples, oranges, peaches, tomatoes" | split: ", "',
                AssignmentExpression(
                    name="fruits",
                    expression=FilteredExpression(
                        expression=StringLiteral(
                            value="apples, oranges, peaches, tomatoes",
                        ),
                        filters=[Filter(name="split", args=[StringLiteral(", ")])],
                    ),
                ),
            ),
            Case(
                "assign with multiple filters and identifier arguments",
                "everything = fruits | concat: vegetables | concat: furniture",
                AssignmentExpression(
                    name="everything",
                    expression=FilteredExpression(
                        expression=Identifier(
                            path=[IdentifierPathElement("fruits")],
                        ),
                        filters=[
                            Filter(
                                name="concat",
                                args=[
                                    Identifier(
                                        path=[IdentifierPathElement("vegetables")],
                                    )
                                ],
                            ),
                            Filter(
                                name="concat",
                                args=[
                                    Identifier(
                                        path=[IdentifierPathElement("furniture")],
                                    )
                                ],
                            ),
                        ],
                    ),
                ),
            ),
        ]

        self._test(
            test_cases, tokenize_assignment_expression, parse_assignment_expression
        )

    def test_parse_loop_expression(self):
        """Test that we can parse loop expressions."""
        test_cases = [
            Case(
                "item in array",
                "product in collection.products",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                ),
            ),
            Case(
                "item in array with integer literal limit",
                "product in collection.products limit:5",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                ),
            ),
            Case(
                "item in array with identifier limit",
                "product in collection.products limit:max",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=Identifier(path=[IdentifierPathElement("max")]),
                ),
            ),
            Case(
                "item in array with integer literal offset",
                "product in collection.products offset:2",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array with integer literal offset and limit",
                "product in collection.products limit:5 offset:2",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array with integer literal offset and limit out of order",
                "product in collection.products offset:2 limit:5",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                ),
            ),
            Case(
                "item in array reversed",
                "product in collection.products reversed",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    reversed_=True,
                ),
            ),
            Case(
                "item in array with all options",
                "product in collection.products offset:2 limit:5 reversed",
                LoopExpression(
                    name="product",
                    identifier=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                        ],
                    ),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                    reversed_=True,
                ),
            ),
            Case(
                "item in range with integer literal start and stop",
                "product in (1..10)",
                LoopExpression(
                    name="product",
                    start=IntegerLiteral(1),
                    stop=IntegerLiteral(10),
                ),
            ),
            Case(
                "item in range with integer literal start and identifier stop",
                "product in (1..collection.products.size)",
                LoopExpression(
                    name="product",
                    start=IntegerLiteral(1),
                    stop=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement("size"),
                        ],
                    ),
                ),
            ),
            Case(
                "item in range with integer identifier start and stop",
                "product in (num..collection.products.size)",
                LoopExpression(
                    name="product",
                    start=Identifier(path=[IdentifierPathElement("num")]),
                    stop=Identifier(
                        path=[
                            IdentifierPathElement("collection"),
                            IdentifierPathElement("products"),
                            IdentifierPathElement("size"),
                        ],
                    ),
                ),
            ),
            Case(
                "item in range with options",
                "product in (1..10) offset:2 limit:5 reversed",
                LoopExpression(
                    name="product",
                    start=IntegerLiteral(1),
                    stop=IntegerLiteral(10),
                    limit=IntegerLiteral(5),
                    offset=IntegerLiteral(2),
                    reversed_=True,
                ),
            ),
        ]

        self._test(test_cases, tokenize_loop_expression, parse_loop_expression)



"""Filter functions that can operate on multiple input data types."""

import datetime
import functools

from dateutil import parser

from liquid.filter import Filter
from liquid.exceptions import FilterArgumentError
from liquid.expression import EMPTY


class AbstractFilter(Filter):

    name = "AbstractStringFilter"
    num_args = 0
    msg = "{}: unexpected argument type {}"

    def __call__(self, val, *args, **kwargs):
        if len(args) != self.num_args:
            raise FilterArgumentError(
                f"{self.name}: expected {self.num_args} arguments, found {len(args)}"
            )
        try:
            return self.filter(val, *args, **kwargs)
        except (AttributeError, TypeError) as err:
            raise FilterArgumentError(self.msg.format(self.name, type(val))) from err

    def filter(self, val, *args, **kwargs):
        raise NotImplementedError(":(")


class Size(AbstractFilter):
    """Return the length of an array or string."""

    __slots__ = ()

    name = "size"
    msg = "{}: expected an array or string, found {}"

    def filter(self, obj):
        return len(obj)


class Default(AbstractFilter):
    """Return a default value if the input is nil, false, or empty."""

    __slots__ = ()

    name = "default"
    num_args = 1

    def filter(self, obj, default, allow_false=None):
        if allow_false is True and obj is False:
            return obj

        if obj in (None, False, EMPTY):
            return default

        return obj


class Date(AbstractFilter):
    """Converts a timestamp into another date format."""

    __slots__ = ()

    name = "date"
    num_args = 1

    @functools.lru_cache(maxsize=10)
    def filter(self, dt, fmt):
        if isinstance(dt, str):
            if dt == "now":
                dt = datetime.datetime.now()
            else:
                dt = parser.parse(dt)

        if not isinstance(dt, (datetime.datetime, datetime.date)):
            raise FilterArgumentError(
                f"{self.name}: expected datetime.datetime, found {type(dt)} "
            )

        # TODO: Compare Ruby and Python strftime format codes.
        return dt.strftime(fmt)



"""Pseudo tag and node definition for template literals."""

from typing import TextIO

from liquid.ast import Node
from liquid.context import Context
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_LITERAL


class LiteralNode(Node):
    """Parse tree node for template literals."""

    __slots__ = ("tok",)

    def __init__(self, tok: Token):
        self.tok = tok

    def __str__(self) -> str:
        return self.tok.value

    def __repr__(self) -> str:  # pragma: no cover
        return f"LiteralNode(tok={self.tok})"

    def render_to_output(self, context: Context, buffer: TextIO):
        buffer.write(self.tok.value)


class Literal(Tag):
    """Pseudo "tag" for template literals."""

    name = TOKEN_LITERAL

    def parse(self, stream: TokenStream) -> LiteralNode:
        return LiteralNode(stream.current)



def fnv1a32(val: bytes) -> int:
    hash_ = 0x811C9DC5

    for byte in val:
        hash_ = hash_ ^ byte
        hash_ = (hash_ * 0x01000193) % 0x100000000

    return hash_


def fnv1a16(val: bytes) -> int:
    hash_ = fnv1a32(val)
    return (hash_ >> 16) ^ (hash_ & 0xFFFF)


def fnv(val: str) -> int:
    return fnv1a16(val.encode())



import collections.abc
from abc import ABC, abstractmethod
from typing import Any


class IterableDrop(ABC, collections.abc.Collection):
    """Inherit from this class if you want the drop to be iterable from
    tags like 'include' and 'render'."""

    @abstractmethod
    def step(self, item: Any):
        """Step the iterator."""



"""Liquid render tests."""

import datetime
import unittest

from dataclasses import dataclass
from dataclasses import field

from typing import Mapping
from typing import Any

from liquid.environment import Environment
from liquid.template import AwareBoundTemplate
from liquid.mode import Mode
from liquid.loaders import DictLoader
from liquid.builtin.drops import IterableDrop


class MockIterableDrop(IterableDrop):
    """Mock implementation of an iterable drop."""

    def __init__(self):
        self.items = [
            {"foo": 1, "bar": 2},
            {"foo": 2, "bar": 1},
            {"foo": 3, "bar": 3},
        ]

        self.it = iter(self)

    def __contains__(self, item):
        return False

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def __str__(self):
        return "MockIterableDrop"

    def step(self, item):
        return next(self.it)


TEMPLATE_DROP_ATTRS = (
    r"{{ template }} "
    r"{{ template.directory }} "
    r"{{ template.name }} "
    r"{{ template.suffix }}"
)


@dataclass
class Case:
    """Test case dataclass to help with table driven tests."""

    description: str
    template: str
    expect: str
    globals: Mapping[str, Any] = field(default_factory=dict)
    partials: Mapping[str, Any] = field(default_factory=dict)


class RenderTestCases(unittest.TestCase):
    """Test cases for testing template renders."""

    def _test(self, test_cases, template_class=AwareBoundTemplate):
        """Helper method for testing lists of test cases."""
        self.maxDiff = None

        for case in test_cases:
            env = Environment(loader=DictLoader(case.partials))
            env.template_class = template_class

            template = env.from_string(case.template, globals=case.globals)

            with self.subTest(msg=case.description):
                result = template.render()
                self.assertEqual(result, case.expect)

    def test_literal(self):
        """Test that we can render template literals."""

        self._test(
            [
                Case(
                    description="a literal string",
                    template="a literal string",
                    expect="a literal string",
                ),
                Case(
                    description="some css",
                    template=" div { font-weight: bold; } ",
                    expect=" div { font-weight: bold; } ",
                ),
            ]
        )

    def test_whitespace_control(self):
        """Test that we can control whitespace."""

        self._test(
            [
                Case(
                    description="mixed whitespace control",
                    template="".join(
                        [
                            "\n{% if customer -%}\n",
                            "Welcome back, {{ customer.first_name -}} !\n",
                            "{%- endif -%}",
                        ]
                    ),
                    expect="\nWelcome back, Holly!",
                    globals={"customer": {"first_name": "Holly"}},
                ),
            ]
        )

    def test_output_statement(self):
        """Test that we can render output statements."""

        self._test(
            [
                Case(
                    description="string literal",
                    template=r"{{ 'hello' }}",
                    expect="hello",
                ),
                Case(
                    description="integer literal",
                    template=r"{{ 123 }}",
                    expect="123",
                ),
                Case(
                    description="float literal",
                    template=r"{{ 1.23 }}",
                    expect="1.23",
                ),
                Case(
                    description="global identifier",
                    template=r"{{ product.title }}",
                    expect="foo",
                    globals={"product": {"title": "foo"}},
                ),
                Case(
                    description="local identifier",
                    template=r"{% assign name = 'Brian' %}{{ name }}",
                    expect="Brian",
                ),
                Case(
                    description="global identifier does not exist",
                    template=r"{{ age }}",
                    expect="",
                ),
                Case(
                    description="local identifier does not exist",
                    template=r"{{ age }}",
                    expect="",
                ),
                Case(
                    description="array item by index",
                    template=r"{{ product.tags[1] }}",
                    expect="garden",
                    globals={"product": {"tags": ["sports", "garden"]}},
                ),
                Case(
                    description="array item by index from identifier",
                    template="{% assign i = 1 %}{{ product.tags[i] }}",
                    expect="garden",
                    globals={"product": {"tags": ["sports", "garden"]}},
                ),
                Case(
                    description="global identifier with a filter",
                    template=r"{{ product.title | upcase }}",
                    expect="FOO",
                    globals={"product": {"title": "foo"}},
                ),
                Case(
                    description="dump an array from the global context",
                    template=r"{{ product.tags }}",
                    expect="sportsgarden",
                    globals={"product": {"tags": ["sports", "garden"]}},
                ),
                Case(
                    description="assign by value",
                    template=(
                        r"{% capture some %}hello{% endcapture %}"
                        r"{% assign other = some %}"
                        r"{% assign some = 'foo' %}"
                        r"{{ some }}-{{ other }}"
                    ),
                    expect="foo-hello",
                ),
                Case(
                    description="traverse variables with bracketed identifiers",
                    template=r"{{ site.data.menu[include.menu][include.locale] }}",
                    expect="it works!",
                    globals={
                        "site": {"data": {"menu": {"foo": {"bar": "it works!"}}}},
                        "include": {"menu": "foo", "locale": "bar"},
                    },
                ),
                Case(
                    description="special built in variable 'now'",
                    template=r"{{ now | date: '%d/%m/%Y' }}",
                    expect=datetime.datetime.now().strftime(r"%d/%m/%Y"),
                ),
                Case(
                    description="special built in variable 'today'",
                    template=r"{{ today | date: '%d/%m/%Y' }}",
                    expect=datetime.date.today().strftime(r"%d/%m/%Y"),
                ),
                Case(
                    description="default filter on false literal",
                    template=r"{{ false | default: 'bar' }}",
                    expect="bar",
                ),
                Case(
                    description="default filter allow false",
                    template=r"{{ false | default: 'bar', allow_false: true }}",
                    expect="false",
                ),
                Case(
                    description="default filter don't allow false",
                    template=r"{{ false | default: 'bar', allow_false: false }}",
                    expect="bar",
                ),
            ]
        )

    def test_echo_tag(self):
        """Test that we can render `echo` tags."""

        self._test(
            [
                Case(
                    description="string literal",
                    template=r"{% echo 'hello' %}",
                    expect="hello",
                ),
                Case(
                    description="integer literal",
                    template=r"{% echo 123 %}",
                    expect="123",
                ),
                Case(
                    description="float literal",
                    template=r"{% echo 1.23 %}",
                    expect="1.23",
                ),
                Case(
                    description="global identifier",
                    template=r"{% echo product.title %}",
                    expect="foo",
                    globals={"product": {"title": "foo"}},
                ),
                Case(
                    description="local identifier",
                    template=r"{% assign name = 'Brian' %}{% echo name %}",
                    expect="Brian",
                ),
                Case(
                    description="global identifier does not exist",
                    template=r"{% echo age %}",
                    expect="",
                ),
                Case(
                    description="local identifier does not exist",
                    template=r"{% echo age %}",
                    expect="",
                ),
                Case(
                    description="array item by index",
                    template=r"{% echo product.tags[1] %}",
                    expect="garden",
                    globals={"product": {"tags": ["sports", "garden"]}},
                ),
                Case(
                    description="array item by index from identifier",
                    template=r"{% assign i = 1 %}{% echo product.tags[i] %}",
                    expect="garden",
                    globals={"product": {"tags": ["sports", "garden"]}},
                ),
                Case(
                    description="global identifier with a filter",
                    template=r"{% echo product.title | upcase %}",
                    expect="FOO",
                    globals={"product": {"title": "foo"}},
                ),
                Case(
                    description="dump an array from the global context",
                    template=r"{% echo product.tags %}",
                    expect="sportsgarden",
                    globals={"product": {"tags": ["sports", "garden"]}},
                ),
                Case(
                    description="assign by value",
                    template=(
                        r"{% capture some %}hello{% endcapture %}"
                        r"{% assign other = some %}"
                        r"{% assign some = 'foo' %}"
                        r"{% echo some %}-{% echo other %}"
                    ),
                    expect="foo-hello",
                ),
                Case(
                    description="traverse variables with bracketed identifiers",
                    template=r"{% echo site.data.menu[include.menu][include.locale] %}",
                    expect="it works!",
                    globals={
                        "site": {"data": {"menu": {"foo": {"bar": "it works!"}}}},
                        "include": {"menu": "foo", "locale": "bar"},
                    },
                ),
                Case(
                    description="special built in variable 'now'",
                    template=r"{% echo now | date: '%d/%m/%Y' %}",
                    expect=datetime.datetime.now().strftime(r"%d/%m/%Y"),
                ),
                Case(
                    description="special built in variable 'today'",
                    template=r"{% echo today | date: '%d/%m/%Y' %}",
                    expect=datetime.date.today().strftime(r"%d/%m/%Y"),
                ),
            ]
        )

    def test_assign_tag(self):
        """Test that we can render assigned variables."""
        self._test(
            [
                Case(
                    description="assing a filtered literal",
                    template="{% assign foo = 'foo' | upcase %}{{ foo }}",
                    expect="FOO",
                ),
            ]
        )

    def test_if_tag(self):
        """Test that we can render `if` tags."""
        test_cases = [
            Case(
                description="condition with literal consequence",
                template=r"{% if product.title == 'foo' %}bar{% endif %}",
                expect="bar",
                globals={"product": {"title": "foo"}},
            ),
            Case(
                description="condition with literal consequence and literal alternative",
                template=r"{% if product.title == 'hello' %}bar{% else %}baz{% endif %}",
                expect="baz",
                globals={"product": {"title": "foo"}},
            ),
            Case(
                description="condition with conditional alternative",
                template=(
                    r"{% if product.title == 'hello' %}"
                    r"foo"
                    r"{% elsif product.title == 'foo' %}"
                    r"bar"
                    r"{% endif %}"
                ),
                expect="bar",
                globals={"product": {"title": "foo"}},
            ),
            Case(
                description="condition with conditional alternative and final alternative",
                template=(
                    r"{% if product.title == 'hello' %}"
                    r"foo"
                    r"{% elsif product.title == 'goodbye' %}"
                    r"bar"
                    r"{% else %}"
                    r"baz"
                    r"{% endif %}"
                ),
                expect="baz",
                globals={"product": {"title": "foo"}},
            ),
            Case(
                description="truthy-ness of a dictionary",
                template=r"{% if product %}bar{% else %}foo{% endif %}",
                expect="bar",
                globals={"product": {"title": "foo"}},
            ),
            Case(
                description="falsey-ness of a literal nil",
                template=r"{% if nil %}bar{% else %}foo{% endif %}",
                expect="foo",
            ),
            Case(
                description="falsey-ness of a non existant name",
                template=r"{% if nosuchthing %}bar{% else %}foo{% endif %}",
                expect="foo",
            ),
            Case(
                description="nested condition in the consequence block",
                template=r"{% if product %}{% if title == 'Hello' %}baz{% endif %}{% endif %}",
                expect="baz",
                globals={
                    "product": {"title": "foo"},
                    "title": "Hello",
                },
            ),
            Case(
                description="nested condition, alternative in the consequence block",
                template=(
                    r"{% if product %}"
                    r"{% if title == 'goodbye' %}"
                    r"baz"
                    r"{% else %}"
                    r"hello"
                    r"{% endif %}"
                    r"{% endif %}"
                ),
                expect="hello",
                globals={"product": {"title": "foo"}, "title": "Hello"},
            ),
            Case(
                description="false",
                template=r"{% if false %}{% endif %}",
                expect="",
            ),
            Case(
                description="contains condition",
                template=r"{% if product.tags contains 'garden' %}baz{% endif %}",
                expect="baz",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="not equal condition",
                template=r"{% if product.title != 'foo' %}baz{% endif %}",
                expect="",
                globals={"product": {"title": "foo"}},
            ),
            Case(
                description="alternate not equal condition",
                template=r"{% if product.title <> 'foo' %}baz{% endif %}",
                expect="",
                globals={"product": {"title": "foo"}},
            ),
            Case(
                description="blank empty 'if'",
                template=r"{% if true %}  {% elsif false %} {% else %} {% endif %}",
                expect="",
            ),
            Case(
                description="blank nested block",
                template=r"{% if true %} {% comment %} this is blank {% endcomment %} {% endif %}",
                expect="",
            ),
            Case(
                description="compare to blank",
                template=r"{% if username == blank %}username is blank{% endif %}",
                expect="username is blank",
                globals={"username": ""},
            ),
        ]

        self._test(test_cases)

    def test_comment_tag(self):
        """Test that we can render comment tags."""

        self._test(
            [
                Case(
                    description="simple comment",
                    template=r"{% comment %}foo{% endcomment %}",
                    expect="",
                ),
                Case(
                    description="comment with tags",
                    template=(
                        r"{% comment %}"
                        r"{% if true %}"
                        r"{{ title }}"
                        r"{% endif %}"
                        r"{% endcomment %}"
                    ),
                    expect="",
                ),
            ]
        )

    def test_unless_tag(self):
        """Test that we can render `unless` tags."""

        test_cases = [
            Case(
                description="false",
                template=r"{% unless false %}foo{% endunless %}",
                expect="foo",
            ),
            Case(
                description="true",
                template=r"{% unless true %}foo{% endunless %}",
                expect="",
            ),
            Case(
                description="blank empty block",
                template=r"{% unless false %}  {% endunless %}",
                expect="",
            ),
        ]

        self._test(test_cases)

    def test_capture_tag(self):
        """Test that we can render `capture` tags."""

        test_cases = [
            Case(
                description="literal and global variable",
                template=(
                    r"{% capture greeting %}"
                    r"Hello, {{ customer.first_name }}."
                    r"{% endcapture %}"
                    r"{{ greeting }}"
                ),
                expect="Hello, Holly.",
                globals={"customer": {"first_name": "Holly"}},
            ),
            Case(
                description="hyphen in variable name",
                template=(
                    r"{% capture this-thing %}"
                    r"Hello, {{ customer.first_name }}."
                    r"{% endcapture %}"
                    r"{{ this-thing }}"
                ),
                expect="Hello, Holly.",
                globals={"customer": {"first_name": "Holly"}},
            ),
            Case(
                description="assign from capture",
                template=(
                    r"{% capture some %}"
                    r"hello"
                    r"{% endcapture %}"
                    r"{% assign other = some %}"
                    r"{{ some }}-{{ other }}"
                ),
                expect="hello-hello",
            ),
        ]

        self._test(test_cases)

    def test_case_tag(self):
        """Test that we can render `case` tags."""

        test_cases = [
            Case(
                description="simple case",
                template=(
                    r"{% case title %}"
                    r"{% when 'foo' %}foo"
                    r"{% when 'Hello' %}bar"
                    r"{% endcase %}"
                ),
                expect="bar",
                globals={"title": "Hello"},
            ),
            Case(
                description="identifier 'when' expression ",
                template=(
                    r"{% case title %}"
                    r"{% when other %}foo"
                    r"{% when 'Hello' %}bar"
                    r"{% endcase %}"
                ),
                expect="foo",
                globals={"title": "Hello", "other": "Hello"},
            ),
            Case(
                description="out of scope identifier 'when' expression ",
                template=(
                    r"{% case title %}"
                    r"{% when nosuchthing %}foo"
                    r"{% when 'Hello' %}bar"
                    r"{% endcase %}"
                ),
                expect="bar",
                globals={"title": "Hello"},
            ),
            Case(
                description="name not in scope",
                template=(
                    r"{% case nosuchthing %}"
                    r"{% when 'foo' %}foo"
                    r"{% when 'bar' %}bar"
                    r"{% endcase %}"
                ),
                expect="",
            ),
            Case(
                description="no match and no default",
                template=(
                    r"{% case title %}"
                    r"{% when 'foo' %}foo"
                    r"{% when 'bar' %}bar"
                    r"{% endcase %}"
                ),
                expect="",
                globals={"title": "Hello"},
            ),
            Case(
                description="with default",
                template=(
                    r"{% case title %}"
                    r"{% when 'foo' %}foo"
                    r"{% else %}bar"
                    r"{% endcase %}"
                ),
                expect="bar",
                globals={"title": "Hello"},
            ),
            Case(
                description="no whens",
                template=r"{% case title %}{% else %}bar{% endcase %}",
                expect="bar",
                globals={"title": "Hello"},
            ),
            Case(
                description="no whens or default",
                template=r"{% case title %}{% endcase %}",
                expect="",
                globals={"title": "Hello"},
            ),
            Case(
                description="whitespace",
                template=(
                    "{% case title %}  \n\t"
                    "{% when 'foo' %}foo\n"
                    "{% when 'Hello' %}bar"
                    "{% endcase %}"
                ),
                expect="bar",
                globals={"title": "Hello"},
            ),
        ]

        self._test(test_cases)

    def test_cycle_tag(self):
        """Test that we can render `cycle` tags."""

        test_cases = [
            Case(
                description="no identifier",
                template=(
                    r"{% cycle 'some', 'other' %}"
                    r"{% cycle 'some', 'other' %}"
                    r"{% cycle 'some', 'other' %}"
                ),
                expect="someothersome",
            ),
            Case(
                description="with identifier",
                template=(
                    r"{% cycle 'foo': 'some', 'other' %}"
                    r"{% cycle 'some', 'other' %}"
                    r"{% cycle 'foo': 'some', 'other' %}"
                ),
                expect="somesomeother",
            ),
            Case(
                description="different items",
                template=(
                    r"{% cycle '1', '2', '3' %}"
                    r"{% cycle '1', '2' %}"
                    r"{% cycle '1', '2', '3' %}"
                ),
                expect="112",
            ),
            Case(
                description="integers",
                template=(r"{% cycle 1, 2, 3 %}{% cycle 1, 2, 3 %}{% cycle 1, 2, 3 %}"),
                expect="123",
            ),
        ]

        self._test(test_cases)

    def test_decrement_tag(self):
        """Test that we can render `decrement` tags."""

        test_cases = [
            Case(
                description="named counter",
                template=r"{% decrement foo %}{{ foo }} {% decrement foo %}{{ foo }}",
                expect="-1 -2",
            ),
            Case(
                description="increment and decrement named counter",
                template=r"{% decrement foo %} {% decrement foo %} {% increment foo %}",
                expect="-1 -2 -1",
            ),
        ]

        self._test(test_cases)

    def test_increment_tag(self):
        """Test that we can render `increment` tags."""

        test_cases = [
            Case(
                description="named counter",
                template=r"{% increment foo %} {% increment foo %} {% increment foo %}",
                expect="0 1 2",
            ),
            Case(
                description="multiple named counters",
                template=(
                    r"{% increment foo %} "
                    r"{% increment bar %} "
                    r"{% increment foo %} "
                    r"{% increment bar %}"
                ),
                expect="0 0 1 1",
            ),
            Case(
                description="assign and increment",
                template=(
                    r"{% assign foo = 5 %}"
                    r"{{ foo }} "
                    r"{% increment foo %} "
                    r"{% increment foo %} "
                    r"{{ foo }}"
                ),
                expect="5 0 1 5",
            ),
        ]

        self._test(test_cases)

    def test_for_tag(self):
        """Test that we can render `for` tags."""

        test_cases = [
            Case(
                description="simple range loop",
                template=r"{% for i in (0..3) %}{{ i }} {% endfor %}",
                expect="0 1 2 3 ",
            ),
            Case(
                description="range loop using identifier",
                template=(
                    r"{% for i in (0..product.end_range) %}"
                    r"{{ i }} - {{ product.tags[i] }} "
                    r"{% endfor %}"
                ),
                expect="0 - sports 1 - garden ",
                globals={"product": {"tags": ["sports", "garden"], "end_range": 1}},
            ),
            Case(
                description="simple array loop",
                template=r"{% for tag in product.tags %}{{ tag }} {% endfor %}",
                expect="sports garden ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="simple hash loop",
                template=r"{% for c in collection %}{{ c[0] }} {{ c[1] }} {% endfor %}",
                expect="title foo description bar ",
                globals={"collection": {"title": "foo", "description": "bar"}},
            ),
            Case(
                description="empty array with default",
                template=(
                    r"{% for img in emptythings.array %}"
                    r"{{ img.url }} "
                    r"{% else %}"
                    r"no images"
                    r"{% endfor %}"
                ),
                expect="no images",
                globals={"emptythings": {"array": [], "map": {}, "string": ""}},
            ),
            Case(
                description="break",
                template=(
                    r"{% for tag in product.tags %}"
                    r"{% if tag == 'sports' %}"
                    r"{% break %}"
                    r"{% else %}"
                    r"{{ tag }} "
                    r"{% endif %}"
                    r"{% else %}"
                    r"no images"
                    r"{% endfor %}"
                ),
                expect="",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="continue",
                template=(
                    r"{% for tag in product.tags %}"
                    r"{% if tag == 'sports' %}"
                    r"{% continue %}"
                    r"{% else %}"
                    r"{{ tag }} "
                    r"{% endif %}"
                    r"{% else %}"
                    r"no images"
                    r"{% endfor %}"
                ),
                expect="garden ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="limit",
                template=r"{% for tag in product.tags limit:1 %}{{ tag }} {% endfor %}",
                expect="sports ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="offset",
                template=r"{% for tag in product.tags offset:1 %}{{ tag }} {% endfor %}",
                expect="garden ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop length",
                template=r"{% for tag in product.tags %}{{ forloop.length }} {% endfor %}",
                expect="2 2 ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop length with limit",
                template=r"{% for tag in product.tags limit:3 %}{{ forloop.length }} {% endfor %}",
                expect="3 3 3 ",
                globals={
                    "product": {
                        "tags": [
                            "sports",
                            "garden",
                            "sports",
                            "garden",
                            "sports",
                            "garden",
                        ]
                    }
                },
            ),
            Case(
                description="forloop length with offset",
                template=r"{% for tag in product.tags offset:3 %}{{ forloop.length }} {% endfor %}",
                expect="3 3 3 ",
                globals={
                    "product": {
                        "tags": [
                            "sports",
                            "garden",
                            "sports",
                            "garden",
                            "sports",
                            "garden",
                        ]
                    }
                },
            ),
            Case(
                description="forloop goes out of scope",
                template=(
                    r"{% for tag in product.tags %}"
                    r"{{ forloop.length }} "
                    r"{% endfor %}"
                    r"{{ forloop.length }}"
                ),
                expect="2 2 ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop.first",
                template=r"{% for tag in product.tags %}{{ forloop.first }} {% endfor %}",
                expect="true false ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop.last",
                template=r"{% for tag in product.tags %}{{ forloop.last }} {% endfor %}",
                expect="false true ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop.index",
                template=r"{% for tag in product.tags %}{{ forloop.index }} {% endfor %}",
                expect="1 2 ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop.index0",
                template=r"{% for tag in product.tags %}{{ forloop.index0 }} {% endfor %}",
                expect="0 1 ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop.rindex",
                template=r"{% for tag in product.tags %}{{ forloop.rindex }} {% endfor %}",
                expect="2 1 ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop.rindex0",
                template=r"{% for tag in product.tags %}{{ forloop.rindex0 }} {% endfor %}",
                expect="1 0 ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="forloop no such attribute",
                template=r"{% for tag in product.tags %}{{ forloop.nosuchthing }}{% endfor %}",
                expect="",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="iterate an empty array",
                template=r"{% for item in emptythings.array %}{{ item }}{% endfor %}",
                expect="",
                globals={"emptythings": {"array": [], "map": {}, "string": ""}},
            ),
            Case(
                description="iterate an empty array with default",
                template=r"{% for item in emptythings.array %}{{ item }}{% else %}foo{% endfor %}",
                expect="foo",
                globals={"emptythings": {"array": [], "map": {}, "string": ""}},
            ),
            Case(
                description="lookup a filter from an outer context",
                template=r"{% for tag in product.tags %}{{ tag | upcase }} {% endfor %}",
                expect="SPORTS GARDEN ",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="assign inside loop",
                template=r"{% for tag in product.tags %}{% assign x = tag %}{% endfor %}{{ x }}",
                expect="garden",
                globals={"product": {"tags": ["sports", "garden"]}},
            ),
            Case(
                description="blank empty loops",
                template=r"{% for i in (0..10) %}  {% endfor %}",
                expect="",
            ),
            Case(
                description=(
                    "loop over nested and chained object from context "
                    "with trailing identifier"
                ),
                template=r"{% for link in linklists[section.settings.menu].links %}{{ link }} {% endfor %}",
                expect="1 2 ",
                globals={
                    "linklists": {"main": {"links": ["1", "2"]}},
                    "section": {"settings": {"menu": "main"}},
                },
            ),
        ]

        self._test(test_cases)

    def test_raw_tag(self):
        """Test that we can render `raw` tags."""

        test_cases = [
            Case(
                description="literal",
                template=r"{% raw %}foo{% endraw %}",
                expect="foo",
            ),
            Case(
                description="statement",
                template=r"{% raw %}{{ foo }}{% endraw %}",
                expect=r"{{ foo }}",
            ),
            Case(
                description="tag",
                template=r"{% raw %}{% assign x = 1 %}{% endraw %}",
                expect=r"{% assign x = 1 %}",
            ),
            Case(
                description="partial tag",
                template=r"{% raw %} %} {% }} {{ {% endraw %}",
                expect=r" %} {% }} {{ ",
            ),
            Case(
                description="continue after raw",
                template=r"{% raw %} {% some raw content %} {% endraw %}a literal",
                expect=r" {% some raw content %} a literal",
            ),
        ]

        self._test(test_cases)

    def test_tablerow_tag(self):
        """Test that we can render `tablerow` tags."""

        test_cases = [
            Case(
                description="one row",
                template=r"{% tablerow tag in collection.tags %}{{ tag }}{% endtablerow %}",
                expect=(
                    '<tr class="row1">'
                    '<td class="col1">tag1</td>'
                    '<td class="col2">tag2</td>'
                    '<td class="col3">tag3</td>'
                    '<td class="col4">tag4</td>'
                    "</tr>"
                ),
                globals={
                    "collection": {
                        "tags": [
                            "tag1",
                            "tag2",
                            "tag3",
                            "tag4",
                        ]
                    }
                },
            ),
            Case(
                description="two columns",
                template=r"{% tablerow tag in collection.tags cols:2 %}{{ tag }}{% endtablerow %}",
                expect=(
                    '<tr class="row1">'
                    '<td class="col1">tag1</td>'
                    '<td class="col2">tag2</td>'
                    "</tr>"
                    '<tr class="row2">'
                    '<td class="col1">tag3</td>'
                    '<td class="col2">tag4</td>'
                    "</tr>"
                ),
                globals={
                    "collection": {
                        "tags": [
                            "tag1",
                            "tag2",
                            "tag3",
                            "tag4",
                        ]
                    }
                },
            ),
            Case(
                description="two column range",
                template=(
                    r"{% tablerow i in (1..4) cols:2 %}"
                    r"{{ i }} {{ tablerowloop.col_first }}"
                    r"{% endtablerow %}"
                ),
                expect=(
                    '<tr class="row1">'
                    '<td class="col1">1 true</td>'
                    '<td class="col2">2 false</td>'
                    "</tr>"
                    '<tr class="row2">'
                    '<td class="col1">3 true</td>'
                    '<td class="col2">4 false</td>'
                    "</tr>"
                ),
            ),
            Case(
                description="two column odd range",
                template=(
                    r"{% tablerow i in (1..5) cols:2 %}"
                    r"{{ i }} {{ tablerowloop.col_first }}"
                    r"{% endtablerow %}"
                ),
                expect=(
                    '<tr class="row1">'
                    '<td class="col1">1 true</td>'
                    '<td class="col2">2 false</td>'
                    "</tr>"
                    '<tr class="row2">'
                    '<td class="col1">3 true</td>'
                    '<td class="col2">4 false</td>'
                    "</tr>"
                    '<tr class="row3">'
                    '<td class="col1">5 true</td>'
                    "</tr>"
                ),
            ),
        ]

        self._test(test_cases)

    def test_liquid_tag(self):
        """Test that we can render `liquid` tags."""

        test_cases = [
            Case(
                description="multiple tags",
                template="\n".join(
                    [
                        r"{% liquid",
                        r"if product.title",
                        r"   echo product.title | upcase",
                        r"else",
                        r"   echo 'product-1' | upcase ",
                        r"endif",
                        r"",
                        r"for i in (0..5)",
                        r"   echo i",
                        r"endfor %}",
                    ]
                ),
                expect="FOO012345",
                globals={"product": {"title": "foo"}},
            ),
        ]

        self._test(test_cases)

    def test_illegal(self):
        """Test that we can render an `illegal` token in LAX mode."""
        env = Environment(tolerance=Mode.LAX)
        template = env.from_string(r"{% nosuchthing %}")
        result = template.render()
        self.assertEqual(result, "")

    def test_include_tag(self):
        """Test that we can render `include` tags."""

        test_cases = [
            Case(
                description="string literal name",
                template=r"{% include 'product-hero' %}{{ partial }}",
                expect="foo\n- sports\n- garden\ntruefalse",
                globals={"product": {"title": "foo", "tags": ["sports", "garden"]}},
                partials={
                    "product-hero": (
                        r"{{ product.title }}"
                        "\n"
                        r"{% for tag in product.tags %}"
                        r"- {{ tag }}"
                        "\n"
                        r"{% endfor %}"
                        r"{{ partial }}"
                    ),
                },
            ),
            Case(
                description="name from identifier",
                template=r"{% include snippet %}",
                expect="foo\n- sports\n- garden\ntrue",
                globals={
                    "snippet": "product-hero",
                    "product": {"title": "foo", "tags": ["sports", "garden"]},
                },
                partials={
                    "product-hero": (
                        r"{{ product.title }}"
                        "\n"
                        r"{% for tag in product.tags %}"
                        r"- {{ tag }}"
                        "\n"
                        r"{% endfor %}"
                        r"{{ partial }}"
                    ),
                },
            ),
            Case(
                description="bound variable",
                template=r"{% include 'product-title' with collection.products[1] %}",
                expect="car",
                globals={
                    "collection": {
                        "products": [{"title": "bike"}, {"title": "car"}],
                    }
                },
                partials={
                    "product-title": r"{{ product-title.title }}",
                },
            ),
            Case(
                description="bound variable does not exist",
                template=r"{% include 'product-title' with no.such.thing %}",
                expect="",
                partials={
                    "product-title": r"{{ product-title.title }}",
                },
            ),
            Case(
                description="bound array variable",
                template=r"{% include 'prod' for collection.products %}",
                expect="bikecar",
                globals={
                    "collection": {
                        "products": [{"title": "bike"}, {"title": "car"}],
                    }
                },
                partials={"prod": r"{{ prod.title }}"},
            ),
            Case(
                description="bound variable with alias",
                template=r"{% include 'product-alias' with collection.products[1] as product %}",
                expect="car",
                globals={
                    "collection": {
                        "products": [{"title": "bike"}, {"title": "car"}],
                    }
                },
                partials={"product-alias": r"{{ product.title }}"},
            ),
            Case(
                description="some keyword arguments",
                template=r"{% include 'product-args', foo: 'hello', bar: 'there' %}",
                expect="hello there",
                partials={"product-args": r"{{ foo }} {{ bar }}"},
            ),
            Case(
                description="template drop",
                template=r"{% include 'some/template-attrs.alt.txt' %}",
                expect="template-attrs.alt some template-attrs alt",
                partials={"some/template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="template drop no parent",
                template=r"{% include 'template-attrs.alt.txt' %}",
                expect="template-attrs.alt  template-attrs alt",
                partials={"template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="template drop no suffix",
                template=r"{% include 'some/template-attrs.txt' %}",
                expect="template-attrs some template-attrs ",
                partials={"some/template-attrs.txt": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="template drop no suffix or extension",
                template=r"{% include 'some/template-attrs' %}",
                expect="template-attrs some template-attrs ",
                partials={"some/template-attrs": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="use globals from outer scope",
                template=r"{% include 'outer-scope' %}",
                expect="Hello, Holly",
                globals={"customer": {"first_name": "Holly"}},
                partials={"outer-scope": r"Hello, {{ customer.first_name }}"},
            ),
            Case(
                description="assign persists in outer scope",
                template=r"{% include 'assign-outer-scope' %} {{ last_name }}",
                expect="Hello, Holly Smith",
                globals={"customer": {"first_name": "Holly"}},
                partials={
                    "assign-outer-scope": (
                        r"Hello, {{ customer.first_name }}"
                        r"{% assign last_name = 'Smith' %}"
                    ),
                },
            ),
            Case(
                description="counter from outer scope",
                template=(
                    r"{% increment foo %} "
                    r"{% include 'increment-outer-scope' %} "
                    r"{% increment foo %}"
                ),
                expect="0 1 2",
                partials={"increment-outer-scope": r"{% increment foo %}"},
            ),
            Case(
                description="break from include",
                template=r"{% for tag in product.tags %}{% include 'tag-break' %}{% endfor %}",
                expect="SPORTS",
                globals={"product": {"tags": ["sports", "garden"]}},
                partials={"tag-break": r"{{ tag | upcase }}{% break %}"},
            ),
            Case(
                description="break from nested include",
                template=r"{% for tag in product.tags %}{% include 'tag' %}{% endfor %}",
                expect="SPORTS break!",
                globals={"product": {"tags": ["sports", "garden"]}},
                partials={
                    "tag": r"{{ tag | upcase }}{% include 'break' %}",
                    "break": r" break!{% break %}",
                },
            ),
        ]

        self._test(test_cases)

    def test_render_tag(self):
        """Test that we can render `render` tags."""

        test_cases = [
            Case(
                description="string literal name",
                template=r"{% render 'product-hero' %}{{ partial }}",
                expect="foo\n- sports - garden \ntruefalse",
                globals={"product": {"title": "foo", "tags": ["sports", "garden"]}},
                partials={
                    "product-hero": "\n".join(
                        [
                            r"{{ product.title }}",
                            r"{% for tag in product.tags %}- {{ tag }} {% endfor %}",
                            r"{{ partial }}",
                        ]
                    ),
                },
            ),
            Case(
                description="bound variable",
                template=r"{% render 'product-title' with collection.products[1] %}",
                expect="car",
                globals={
                    "collection": {"products": [{"title": "bike"}, {"title": "car"}]}
                },
                partials={"product-title": r"{{ product-title.title }}"},
            ),
            Case(
                description="bound variable does not exist",
                template=r"{% render 'product-title' with no.such.thing %}",
                expect="",
                globals={},
                partials={"product-title": r"{{ product-title.title }}"},
            ),
            Case(
                description="bound array variable",
                template=r"{% render 'prod' for collection.products %}",
                expect="bikecar",
                globals={
                    "collection": {"products": [{"title": "bike"}, {"title": "car"}]}
                },
                partials={"prod": r"{{ prod.title }}"},
            ),
            Case(
                description="bound variable with alias",
                template=r"{% render 'product-alias' with collection.products[1] as product %}",
                expect="car",
                globals={
                    "collection": {"products": [{"title": "bike"}, {"title": "car"}]}
                },
                partials={
                    "product-alias": r"{{ product.title }}",
                },
            ),
            Case(
                description="some keyword arguments",
                template=r"{% render 'product-args', foo: 'hello', bar: 'there' %}",
                expect="hello there",
                globals={},
                partials={
                    "product-args": r"{{ foo }} {{ bar }}",
                },
            ),
            Case(
                description="some keyword arguments no leading coma",
                template=r"{% render 'product-args' foo: 'hello', bar: 'there' %}",
                expect="hello there",
                globals={},
                partials={"product-args": r"{{ foo }} {{ bar }}"},
            ),
            Case(
                description="template drop",
                template=r"{% render 'some/template-attrs.alt.txt' %}",
                expect="template-attrs.alt some template-attrs alt",
                globals={},
                partials={"some/template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="template drop no parent",
                template=r"{% render 'template-attrs.alt.txt' %}",
                expect="template-attrs.alt  template-attrs alt",
                globals={},
                partials={"template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="template drop no suffix",
                template=r"{% render 'some/template-attrs.txt' %}",
                expect="template-attrs some template-attrs ",
                globals={},
                partials={"some/template-attrs.txt": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="template drop no suffix or extension",
                template=r"{% render 'some/template-attrs' %}",
                expect="template-attrs some template-attrs ",
                globals={},
                partials={"some/template-attrs": TEMPLATE_DROP_ATTRS},
            ),
            Case(
                description="parent variables go out of scope",
                template=(
                    r"{% assign greeting = 'good morning' %}"
                    r"{{ greeting }} "
                    r"{% render 'outer-scope' %}"
                    r"{{ greeting }}"
                ),
                expect="good morning good morning",
                partials={"outer-scope": r"{{ greeting }}"},
            ),
            Case(
                description="for loop variables go out of scope",
                template=(
                    r"{% for i in (1..3) %}"
                    r"{{ i }}"
                    r"{% render 'loop-scope' %}"
                    r"{{ i }}"
                    r"{% endfor %}"
                    r"{{ i }}"
                ),
                expect="112233",
                partials={"loop-scope": r"{{ i }}"},
            ),
            Case(
                description="assigned variables to not leek into outer scope",
                template=r"{% render 'assign-outer-scope' %} {{ last_name }}",
                expect="Hello, Holly ",
                globals={"customer": {"first_name": "Holly"}},
                partials={
                    "assign-outer-scope": (
                        r"Hello, {{ customer.first_name }}"
                        r"{% assign last_name = 'Smith' %}"
                    )
                },
            ),
            Case(
                description="increment is isolated between renders",
                template=(
                    r"{% increment foo %} "
                    r"{% render 'increment' %} "
                    r"{% increment foo %}"
                ),
                expect="0 0 1",
                partials={"increment": r"{% increment foo %}"},
            ),
            Case(
                description="decrement is isolated between renders",
                template=(
                    r"{% decrement foo %} "
                    r"{% render 'decrement' %} "
                    r"{% decrement foo %}"
                ),
                expect="-1 -1 -2",
                partials={"decrement": r"{% decrement foo %}"},
            ),
            Case(
                description="forloop helper",
                template=r"{% render 'product' for collection.products %}",
                expect="Product: bike first index:1 Product: car last index:2 ",
                globals={
                    "collection": {"products": [{"title": "bike"}, {"title": "car"}]}
                },
                partials={
                    "product": (
                        r"Product: {{ product.title }} "
                        r"{% if forloop.first %}first{% endif %}"
                        r"{% if forloop.last %}last{% endif %}"
                        r" index:{{ forloop.index }} "
                    ),
                },
            ),
            Case(
                description="for drop",
                template=r"{% render 'loop-for' for loop as value %}",
                expect="123",
                globals={"loop": MockIterableDrop()},
                partials={
                    "loop-for": r"{{ value.foo }}",
                },
            ),
            Case(
                description="with drop",
                template=r"{% render 'loop-with' with loop as value %}",
                expect="MockIterableDrop",
                globals={"loop": MockIterableDrop()},
                partials={"loop-with": r"{{ value }}"},
            ),
        ]

        self._test(test_cases)

    def test_ifchanged_tag(self):
        """Test that we can render `ifchanged` tags."""

        test_cases = [
            Case(
                description="changed from initial state",
                template=r"{% ifchanged %}hello{% endifchanged %}",
                expect="hello",
            ),
            Case(
                description="not changed from initial state",
                template=r"{% ifchanged %}{% endifchanged %}",
                expect="",
            ),
            Case(
                description="no change from assign",
                template=(
                    r"{% assign foo = 'hello' %}"
                    r"{% ifchanged %}{{ foo }}{% endifchanged %}"
                    r"{% ifchanged %}{{ foo }}{% endifchanged %}"
                ),
                expect="hello",
            ),
            Case(
                description="no change from assign",
                template=(
                    r"{% assign foo = 'hello' %}"
                    r"{% ifchanged %}{{ foo }}{% endifchanged %}"
                    r"{% ifchanged %}{{ foo }}{% endifchanged %}"
                    r"{% assign foo = 'goodbye' %}"
                    r"{% ifchanged %}{{ foo }}{% endifchanged %}"
                ),
                expect="hellogoodbye",
            ),
            Case(
                description="within for loop",
                template=(
                    r'{% assign list = "1,3,2,1,3,1,2" | split: "," | sort %}'
                    r"{% for item in list -%}"
                    r"{%- ifchanged %} {{ item }}{% endifchanged -%}"
                    r"{%- endfor %}"
                ),
                expect=" 1 2 3",
            ),
        ]

        self._test(test_cases)



from liquid.filter import Filter, number_required, no_args

# pylint: disable=arguments-differ, too-few-public-methods


class Weight(Filter):

    __slots__ = ()

    name = "weight"

    @number_required
    @no_args
    def __call__(self, grams):
        return f"{grams / 1000:.2f}"


class WeightWithUnit(Filter):

    __slots__ = ()

    name = "weight_with_unit"

    @number_required
    @no_args
    def __call__(self, grams):
        return f"{grams / 1000:.2f} kg"



"""Tag and node definition for the built-in "if" tag."""
from __future__ import annotations
import sys

from typing import Optional
from typing import List
from typing import TextIO
from typing import TYPE_CHECKING

from liquid.token import Token
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION

from liquid.ast import Node
from liquid.ast import BlockNode
from liquid.ast import ConditionalBlockNode

from liquid.exceptions import LiquidSyntaxError
from liquid.lex import tokenize_boolean_expression

from liquid.parse import expect
from liquid.parse import parse_boolean_expression
from liquid.parse import get_parser
from liquid.parse import eat_block

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.builtin.illegal import IllegalNode

if TYPE_CHECKING:
    from liquid.context import Context
    from liquid import Environment
    from liquid.expression import Expression

TAG_IF = sys.intern("if")
TAG_ENDIF = sys.intern("endif")
TAG_ELSIF = sys.intern("elsif")
TAG_ELSE = sys.intern("else")

ENDIFBLOCK = (TAG_ENDIF, TAG_ELSIF, TAG_ELSE, TOKEN_EOF)
ENDELSIFBLOCK = (TAG_ENDIF, TAG_ELSIF, TAG_ELSE)
ENDIFELSEBLOCK = (TAG_ENDIF,)


class IfNode(Node):
    """Parse tree node for the built-in "if" tag."""

    __slots__ = (
        "tok",
        "condition",
        "consequence",
        "conditional_alternatives",
        "alternative",
    )

    def __init__(
        self,
        tok: Token,
        condition: Expression,
        consequence: BlockNode,
        conditional_alternatives: List[ConditionalBlockNode],
        alternative: Optional[BlockNode],
    ):
        self.tok = tok
        self.condition = condition
        self.consequence = consequence
        self.conditional_alternatives = conditional_alternatives
        self.alternative = alternative

    def __str__(self) -> str:
        buf = [
            f"if {self.condition} {{ {self.consequence} }}",
        ]

        for alt in self.conditional_alternatives:
            buf.append(f"elsif {alt}")

        if self.alternative:
            buf.append(f"else {{ {self.alternative} }}")
        return " ".join(buf)

    def __repr__(self):  # pragma: no cover
        return f"IfNode(tok={self.tok}, condition='{self.condition}')"

    def render_to_output(self, context: Context, buffer: TextIO):
        if self.condition.evaluate(context):
            self.consequence.render(context, buffer)
        else:
            rendered = False
            for alt in self.conditional_alternatives:
                if alt.render(context, buffer):
                    rendered = True
                    break

            if not rendered and self.alternative:
                self.alternative.render(context, buffer)


class IfTag(Tag):
    """The built-in "if" tag."""

    name = TAG_IF
    end = TAG_ENDIF

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse_expression(self, stream: TokenStream) -> Expression:
        expect(stream, TOKEN_EXPRESSION)
        expr_iter = tokenize_boolean_expression(stream.current.value)
        return parse_boolean_expression(TokenStream(expr_iter))

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_IF)
        tok = stream.current
        stream.next_token()

        condition = self.parse_expression(stream)
        stream.next_token()

        consequence = self.parser.parse_block(stream, ENDIFBLOCK)
        conditional_alternatives = []

        while stream.current.istag(TAG_ELSIF):
            stream.next_token()

            # If the expression can't be parsed, eat the "elsif" block and
            # continue to parse more "elsif" expression, if any.
            try:
                expr = self.parse_expression(stream)
            except LiquidSyntaxError as err:
                self.env.error(err)
                eat_block(stream, ENDELSIFBLOCK)
                return IllegalNode(tok)

            alt = ConditionalBlockNode(
                stream.current,
                condition=expr,
            )
            stream.next_token()
            alt.block = self.parser.parse_block(stream, ENDELSIFBLOCK)
            conditional_alternatives.append(alt)

        if stream.current.istag(TAG_ELSE):
            stream.next_token()
            alternative = self.parser.parse_block(stream, ENDIFELSEBLOCK)
        else:
            alternative = None

        expect(stream, TOKEN_TAG, value=TAG_ENDIF)
        return IfNode(
            tok=tok,
            condition=condition,
            consequence=consequence,
            conditional_alternatives=conditional_alternatives,
            alternative=alternative,
        )
